# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import ThreadPoolExecutor

from oslo_utils import timeutils

from blazarclient import base
from blazarclient.i18n import _
from blazarclient import utils


class LeaseClientManager(base.BaseClientManager):
    """Manager for the lease connected requests."""

    def create(self, name, start, end, reservations, events, before_end=None):
        """Creates lease from values passed."""
        values = {'name': name, 'start_date': start, 'end_date': end,
                  'reservations': reservations, 'events': events,
                  'before_end_date': before_end}

        resp, body = self.request_manager.post('/leases', body=values)
        return body['lease']

    def get(self, lease_id, detail=False):
        """Describes lease specifications such as name, status and locked
        condition.
        """
        resp, body = self.request_manager.get('/leases/%s' % lease_id)
        return body['lease']

    def update(self, lease_id, name=None, prolong_for=None, reduce_by=None,
               end_date=None, advance_by=None, defer_by=None, start_date=None,
               reservations=None):
        """Update attributes of the lease."""
        values = {}
        if name:
            values['name'] = name

        lease_end_date_change = prolong_for or reduce_by or end_date
        lease_start_date_change = defer_by or advance_by or start_date
        lease = None

        if lease_end_date_change:
            lease = self.get(lease_id)
            if end_date:
                date = timeutils.parse_strtime(end_date, utils.API_DATE_FORMAT)
                values['end_date'] = date.strftime(utils.API_DATE_FORMAT)
            else:
                self._add_lease_date(values, lease, 'end_date',
                                     lease_end_date_change,
                                     prolong_for is not None)

        if lease_start_date_change:
            if lease is None:
                lease = self.get(lease_id)
            if start_date:
                date = timeutils.parse_strtime(start_date,
                                               utils.API_DATE_FORMAT)
                values['start_date'] = date.strftime(utils.API_DATE_FORMAT)
            else:
                self._add_lease_date(values, lease, 'start_date',
                                     lease_start_date_change,
                                     defer_by is not None)

        if reservations:
            values['reservations'] = reservations

        if not values:
            return _('No values to update passed.')
        resp, body = self.request_manager.put('/leases/%s' % lease_id,
                                              body=values)
        return body['lease']

    def delete(self, lease_id):
        """Deletes lease with specified ID."""
        resp, body = self.request_manager.delete('/leases/%s' % lease_id)

    def list(self, sort_by=None):
        """List all leases."""
        resp, body = self.request_manager.get('/leases')
        leases = body['leases']
        if sort_by:
            leases = sorted(leases, key=lambda lease: lease[sort_by])
        return leases

    def additional_details(self, lease_id):
        allocations = {}
        with ThreadPoolExecutor() as executor:
            # Submit the calls
            h_future = executor.submit(self.hosts_in_lease, lease_id)
            n_future = executor.submit(self.networks_in_lease, lease_id)
            d_future = executor.submit(self.devices_in_lease, lease_id)

            # Retrieve the results
            allocations['hosts'] = h_future.result()
            allocations['networks'] = n_future.result()
            allocations['devices'] = d_future.result()
        return allocations

    def hosts_in_lease(self, lease_id):
        """List all hosts in lease"""
        resp, body = self.request_manager.get(f'/leases/{lease_id}/hosts')
        return body['hosts']

    def networks_in_lease(self, lease_id):
        """List all networks in lease"""
        resp, body = self.request_manager.get(f'/leases/{lease_id}/networks')
        return body['networks']

    def devices_in_lease(self, lease_id):
        """List all devices in lease"""
        resp, body = self.request_manager.get(f'/leases/{lease_id}/devices')
        return body['devices']

    def _add_lease_date(self, values, lease, key, delta_date, positive_delta):
        delta_sec = utils.from_elapsed_time_to_delta(
            delta_date,
            pos_sign=positive_delta)
        date = timeutils.parse_strtime(lease[key],
                                       utils.LEASE_DATE_FORMAT)
        values[key] = (date + delta_sec).strftime(utils.API_DATE_FORMAT)



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import importutils

from blazarclient import client
from blazarclient import exception
from blazarclient import tests


class BaseClientTestCase(tests.TestCase):

    def setUp(self):
        super(BaseClientTestCase, self).setUp()

        self.client = client

        self.import_obj = self.patch(importutils, "import_object")

    def test_with_v1(self):
        self.client.Client()
        self.import_obj.assert_called_once_with(
            'blazarclient.v1.client.Client',
            service_type='reservation')

    def test_with_v1a0(self):
        self.client.Client(version='1a0')
        self.import_obj.assert_called_once_with(
            'blazarclient.v1.client.Client',
            service_type='reservation')

    def test_with_wrong_vers(self):
        self.assertRaises(exception.UnsupportedVersion,
                          self.client.Client,
                          version='0.0')



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazarclient import base
from blazarclient import exception
from blazarclient.i18n import _


class ComputeHostClientManager(base.BaseClientManager):
    """Manager for the ComputeHost connected requests."""

    def create(self, name, **kwargs):
        """Creates host from values passed."""
        values = {'name': name}
        values.update(**kwargs)
        resp, body = self.request_manager.post('/os-hosts', body=values)
        return body['host']

    def get(self, host_id):
        """Describe host specifications such as name and details."""
        resp, body = self.request_manager.get('/os-hosts/%s' % host_id)
        return body['host']

    def update(self, host_id, values):
        """Update attributes of the host."""
        if not values:
            return _('No values to update passed.')
        resp, body = self.request_manager.put(
            '/os-hosts/%s' % host_id, body=values
        )
        return body['host']

    def delete(self, host_id):
        """Delete host with specified ID."""
        resp, body = self.request_manager.delete('/os-hosts/%s' % host_id)

    def list(self, sort_by=None):
        """List all hosts."""
        resp, body = self.request_manager.get('/os-hosts')
        hosts = body['hosts']
        if sort_by:
            hosts = sorted(hosts, key=lambda host: host[sort_by])
        return hosts

    def get_allocation(self, host_id):
        """Get allocation for host."""
        resp, body = self.request_manager.get(
            '/os-hosts/%s/allocation' % host_id)
        return body['allocation']

    def list_allocations(self, sort_by=None):
        """List allocations for all hosts."""
        resp, body = self.request_manager.get('/os-hosts/allocations')
        allocations = body['allocations']
        if sort_by:
            allocations = sorted(allocations, key=lambda alloc: alloc[sort_by])
        return allocations

    def reallocate(self, host_id, values):
        """Reallocate host from leases."""
        resp, body = self.request_manager.put(
            '/os-hosts/%s/allocation' % host_id, body=values)
        return body['allocation']

    def list_properties(self, detail=False, all=False, sort_by=None):
        url = '/os-hosts/properties'

        query_parts = []
        if detail:
            query_parts.append("detail=True")
        if all:
            query_parts.append("all=True")
        if query_parts:
            url += "?" + "&".join(query_parts)

        resp, body = self.request_manager.get(url)
        resource_properties = body['resource_properties']

        # Values is a reserved word in cliff so need to rename values column.
        if detail:
            for p in resource_properties:
                p['property_values'] = p['values']
                del p['values']

        if sort_by:
            resource_properties = sorted(resource_properties,
                                         key=lambda lease: lease[sort_by])
        return resource_properties

    def get_property(self, property_name):
        resource_property = [
            x for x in self.list_properties(detail=True)
            if x['property'] == property_name]
        if not resource_property:
            raise exception.ResourcePropertyNotFound()
        return {} if not resource_property else resource_property[0]

    def set_property(self, property_name, private, is_unique=False):
        data = {'private': private, 'is_unique': is_unique}
        resp, body = self.request_manager.patch(
            '/os-hosts/properties/%s' % property_name, body=data)

        return body['resource_property']



# Copyright (c) 2019 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from unittest import mock

from blazarclient import shell
from blazarclient import tests
from blazarclient.v1.shell_commands import floatingips


class CreateFloatingIPTest(tests.TestCase):

    def setUp(self):
        super(CreateFloatingIPTest, self).setUp()
        self.create_floatingip = floatingips.CreateFloatingIP(
            shell.BlazarShell(), mock.Mock())

    def test_args2body(self):
        args = argparse.Namespace(
            network_id='1e17587e-a7ed-4b82-a17b-4beb32523e28',
            floating_ip_address='172.24.4.101',
        )

        expected = {
            'network_id': '1e17587e-a7ed-4b82-a17b-4beb32523e28',
            'floating_ip_address': '172.24.4.101',
        }

        ret = self.create_floatingip.args2body(args)
        self.assertDictEqual(ret, expected)


class ListFloatingIPsTest(tests.TestCase):

    def create_list_command(self, list_value):
        mock_floatingip_manager = mock.Mock()
        mock_floatingip_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.floatingip = mock_floatingip_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return (floatingips.ListFloatingIPs(blazar_shell, mock.Mock()),
                mock_floatingip_manager)

    def test_list_floatingips(self):
        list_value = [
            {'id': '84c4d37e-1f8b-45ce-897b-16ad7f49b0e9'},
            {'id': 'f180cf4c-f886-4dd1-8c36-854d17fbefb5'},
        ]

        list_floatingips, floatingip_manager = self.create_list_command(
            list_value)

        args = argparse.Namespace(sort_by='id', columns=['id'])
        expected = [['id'], [('84c4d37e-1f8b-45ce-897b-16ad7f49b0e9',),
                             ('f180cf4c-f886-4dd1-8c36-854d17fbefb5',)]]

        ret = list_floatingips.get_data(args)
        # ret is a set
        self.assertEqual(expected[0], list(ret[0]))
        self.assertEqual(expected[1], [x for x in ret[1]])

        floatingip_manager.list.assert_called_once_with(sort_by='id')


class ShowFloatingIPTest(tests.TestCase):

    def create_show_command(self, list_value, get_value):
        mock_floatingip_manager = mock.Mock()
        mock_floatingip_manager.list.return_value = list_value
        mock_floatingip_manager.get.return_value = get_value

        mock_client = mock.Mock()
        mock_client.floatingip = mock_floatingip_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return (floatingips.ShowFloatingIP(blazar_shell, mock.Mock()),
                mock_floatingip_manager)

    def test_show_floatingip(self):
        list_value = [
            {'id': '84c4d37e-1f8b-45ce-897b-16ad7f49b0e9'},
            {'id': 'f180cf4c-f886-4dd1-8c36-854d17fbefb5'},
        ]
        get_value = {
            'id': '84c4d37e-1f8b-45ce-897b-16ad7f49b0e9'}

        show_floatingip, floatingip_manager = self.create_show_command(
            list_value, get_value)

        args = argparse.Namespace(id='84c4d37e-1f8b-45ce-897b-16ad7f49b0e9', formatter="table")
        expected = [('id',), ('84c4d37e-1f8b-45ce-897b-16ad7f49b0e9',)]

        ret = show_floatingip.get_data(args)
        self.assertEqual(ret, expected)

        floatingip_manager.get.assert_called_once_with(
            '84c4d37e-1f8b-45ce-897b-16ad7f49b0e9')


class DeleteFloatingIPTest(tests.TestCase):

    def create_delete_command(self, list_value):
        mock_floatingip_manager = mock.Mock()
        mock_floatingip_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.floatingip = mock_floatingip_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return (floatingips.DeleteFloatingIP(blazar_shell, mock.Mock()),
                mock_floatingip_manager)

    def test_delete_floatingip(self):
        list_value = [
            {'id': '84c4d37e-1f8b-45ce-897b-16ad7f49b0e9'},
            {'id': 'f180cf4c-f886-4dd1-8c36-854d17fbefb5'},
        ]
        delete_floatingip, floatingip_manager = self.create_delete_command(
            list_value)

        args = argparse.Namespace(id='84c4d37e-1f8b-45ce-897b-16ad7f49b0e9')
        delete_floatingip.run(args)

        floatingip_manager.delete.assert_called_once_with(
            '84c4d37e-1f8b-45ce-897b-16ad7f49b0e9')



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from blazarclient import command
from blazarclient import exception

# Matches integers or UUIDs
HOST_ID_PATTERN = r'^([0-9]+|([0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}))$'


class ListHosts(command.ListCommand):
    """Print a list of hosts."""
    resource = 'host'
    log = logging.getLogger(__name__ + '.ListHosts')
    list_columns = ['id', 'hypervisor_hostname', 'vcpus', 'memory_mb',
                    'local_gb']
    long_columns = ['node_name', 'node_type', 'disabled', 'reservable']

    def get_parser(self, prog_name):
        parser = super(ListHosts, self).get_parser(prog_name)
        parser.add_argument(
            '--sort-by', metavar="<host_column>",
            help='column name used to sort result',
            default='hypervisor_hostname'
        )
        parser.add_argument(
            '--node-type', metavar="<node_type>",
            help='Node type to filter leases by',
        )
        parser.add_argument(
            '--reservable',
            help='List only reservable hosts',
            action='store_true',
            default=None,
        )
        parser.add_argument(
            '--unreservable',
            help='List only unreservable hosts',
            action='store_false',
            dest='reservable',
        )
        return parser

    def get_data(self, parsed_args):
        if parsed_args.node_type:
            self._filters.append(
                lambda x: x['node_type'] == parsed_args.node_type
            )
        if parsed_args.reservable is not None:
            self._filters.append(
                lambda x: x['reservable'] == parsed_args.reservable
            )
        return super(ListHosts, self).get_data(parsed_args)


class ShowHost(command.ShowCommand):
    """Show host details."""
    resource = 'host'
    json_indent = 4
    name_key = 'hypervisor_hostname'
    id_pattern = HOST_ID_PATTERN
    log = logging.getLogger(__name__ + '.ShowHost')

    def get_parser(self, prog_name):
        parser = super(ShowHost, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'ID or name of %s to look up'
        else:
            help_str = 'ID of %s to look up'
        parser.add_argument('id', metavar=self.resource.upper(),
                            help=help_str % self.resource)
        return parser


class CreateHost(command.CreateCommand):
    """Create a host."""
    resource = 'host'
    json_indent = 4
    log = logging.getLogger(__name__ + '.CreateHost')

    def get_parser(self, prog_name):
        parser = super(CreateHost, self).get_parser(prog_name)
        parser.add_argument(
            'name', metavar=self.resource.upper(),
            help='Name of the host to add'
        )
        parser.add_argument(
            '--extra', metavar='<key>=<value>',
            action='append',
            dest='extra_capabilities',
            default=[],
            help='Extra capabilities key/value pairs to add for the host'
        )
        return parser

    def args2body(self, parsed_args):
        params = {}
        if parsed_args.name:
            params['name'] = parsed_args.name
        extras = {}
        if parsed_args.extra_capabilities:
            for capa in parsed_args.extra_capabilities:
                key, _sep, value = capa.partition('=')
                # NOTE(sbauza): multiple copies of the same capability will
                #               result in only the last value to be stored
                extras[key] = value
            params.update(extras)
        return params


class UpdateHost(command.UpdateCommand):
    """Update attributes of a host."""
    resource = 'host'
    json_indent = 4
    log = logging.getLogger(__name__ + '.UpdateHost')
    name_key = 'hypervisor_hostname'
    id_pattern = HOST_ID_PATTERN

    def get_parser(self, prog_name):
        parser = super(UpdateHost, self).get_parser(prog_name)
        parser.add_argument(
            '--extra', metavar='<key>=<value>',
            action='append',
            dest='extra_capabilities',
            default=[],
            help='Extra capabilities key/value pairs to update for the host'
        )
        return parser

    def args2body(self, parsed_args):
        params = {}
        extras = {}
        if parsed_args.extra_capabilities:
            for capa in parsed_args.extra_capabilities:
                key, _sep, value = capa.partition('=')
                # NOTE(sbauza): multiple copies of the same capability will
                #               result in only the last value to be stored
                extras[key] = value
            params['values'] = extras
        return params


class UnsetAttributesHost(UpdateHost):
    """Unset attributes of a host."""
    log = logging.getLogger(__name__ + '.UnsetAttributesHost')

    def get_parser(self, prog_name):
        parser = super(UpdateHost, self).get_parser(prog_name)
        parser.add_argument(
            '--extra', metavar='<key>',
            action='append',
            dest='extra_capabilities',
            default=[],
            help='Extra capability keys which should be unset from the host.',
        )
        return parser

    def args2body(self, parsed_args):
        if parsed_args.extra_capabilities:
            return {
                'values': {
                    cap: None for cap in parsed_args.extra_capabilities
                }
            }
        else:
            return {}

class DeleteHost(command.DeleteCommand):
    """Delete a host."""
    resource = 'host'
    log = logging.getLogger(__name__ + '.DeleteHost')
    name_key = 'hypervisor_hostname'
    id_pattern = HOST_ID_PATTERN


class ShowHostAllocation(command.ShowAllocationCommand):
    """Show host allocation details."""
    resource = 'host'
    json_indent = 4
    id_pattern = HOST_ID_PATTERN
    allow_names = False
    log = logging.getLogger(__name__ + '.ShowHostAllocation')


class ListHostAllocations(command.ListAllocationCommand):
    """List host allocations."""
    resource = 'host'
    log = logging.getLogger(__name__ + '.ListHostAllocations')
    list_columns = ['resource_id', 'reservations']

    def get_parser(self, prog_name):
        parser = super(ListHostAllocations, self).get_parser(prog_name)
        parser.add_argument(
            '--sort-by', metavar="<host_column>",
            help='column name used to sort result',
            default='resource_id'
        )
        return parser


class ReallocateHost(command.ReallocateCommand):
    """Reallocate host from current allocations."""
    resource = 'host'
    json_indent = 4
    log = logging.getLogger(__name__ + '.ReallocateHost')
    name_key = 'hypervisor_hostname'
    id_pattern = HOST_ID_PATTERN

    def get_parser(self, prog_name):
        parser = super(ReallocateHost, self).get_parser(prog_name)
        parser.add_argument(
            '--lease-id',
            help='Lease ID to reallocate host from.')
        parser.add_argument(
            '--reservation-id',
            help='Reservation ID to reallocate host from')
        parser.add_argument(
            '--force',
            action='store_true',
            help='Force reallocation off the host, regardless of if an alternate host is found',
            default=False
        )
        return parser

    def args2body(self, parsed_args):
        params = {}

        if parsed_args.reservation_id:
            params['reservation_id'] = parsed_args.reservation_id
        elif parsed_args.lease_id:
            params['lease_id'] = parsed_args.lease_id
        if parsed_args.force:
            params['force'] = parsed_args.force

        return params


class ShowHostProperty(command.ShowPropertyCommand):
    """Show host property."""
    resource = 'host'
    json_indent = 4
    log = logging.getLogger(__name__ + '.ShowHostProperty')


class ListHostProperties(command.ListCommand):
    """List host properties."""
    resource = 'host'
    log = logging.getLogger(__name__ + '.ListHostProperties')
    list_columns = ['property', 'private', 'property_values', 'is_unique']

    def args2body(self, parsed_args):
        params = {
            'detail': parsed_args.detail,
            'all': parsed_args.all,
        }
        if parsed_args.sort_by:
            if parsed_args.sort_by in self.list_columns:
                params['sort_by'] = parsed_args.sort_by
            else:
                msg = 'Invalid sort option %s' % parsed_args.sort_by
                raise exception.BlazarClientException(msg)

        return params

    def retrieve_list(self, parsed_args):
        """Retrieve a list of resources from Blazar server."""
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.list_properties(**body)
        return data

    def get_parser(self, prog_name):
        parser = super(ListHostProperties, self).get_parser(prog_name)
        parser.add_argument(
            '--detail',
            action='store_true',
            help='Return properties with values and attributes.',
            default=False
        )
        parser.add_argument(
            '--sort-by', metavar="<property_column>",
            help='column name used to sort result',
            default='property'
        )
        parser.add_argument(
            '--all',
            action='store_true',
            help='Return all properties, public and private.',
            default=False
        )
        return parser


class UpdateHostProperty(command.UpdatePropertyCommand):
    """Update attributes of a host property."""
    resource = 'host'
    json_indent = 4
    log = logging.getLogger(__name__ + '.UpdateHostProperty')
    name_key = 'property_name'



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazarclient import command
from blazarclient import exception
import logging


class ListDevices(command.ListCommand):
    """Print a list of devices."""
    resource = 'device'
    log = logging.getLogger(__name__ + '.ListDevices')
    list_columns = ['id', 'name', 'device_type']
    long_columns = ['machine_name', 'model', 'device_name', 'device_driver', 'reservable']

    def get_parser(self, prog_name):
        parser = super(ListDevices, self).get_parser(prog_name)
        parser.add_argument(
            '--sort-by', metavar="<device_column>",
            help='column name used to sort result',
            default='name'
        )
        return parser


class ShowDevice(command.ShowCommand):
    """Show device details."""
    resource = 'device'
    json_indent = 4
    name_key = 'name'
    log = logging.getLogger(__name__ + '.ShowDevice')

    def get_parser(self, prog_name):
        parser = super(ShowDevice, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'Name or ID of %s to look up'
        else:
            help_str = 'ID of %s to look up'
        parser.add_argument('id', metavar=self.resource.upper(),
                            help=help_str % self.resource)
        return parser


class CreateDevice(command.CreateCommand):
    """Create a device."""
    resource = 'device'
    json_indent = 4
    log = logging.getLogger(__name__ + '.CreateDevice')

    def get_parser(self, prog_name):
        parser = super(CreateDevice, self).get_parser(prog_name)
        parser.add_argument(
            'name', metavar=self.resource.upper(),
            help='Name of the device to add'
        )
        parser.add_argument(
            '--device-type',
            default='container',
            help='Choose from container, vm, or shell. '
                 'Default to container'
        )
        parser.add_argument(
            '--device-driver',
            default='zun',
            help='The driver of the device'
        )
        parser.add_argument(
            '--extra', metavar='<key>=<value>',
            action='append',
            dest='extra_capabilities',
            default=[],
            help='Extra capabilities key/value pairs to add for the device'
        )
        return parser

    def args2body(self, parsed_args):
        params = {}
        parsed_args_dict = vars(parsed_args)
        for arg in ['name', 'device_type', 'device_driver']:
            if parsed_args_dict.get(arg):
                params[arg] = parsed_args_dict.get(arg)
        extras = {}
        if parsed_args.extra_capabilities:
            for capa in parsed_args.extra_capabilities:
                key, _sep, value = capa.partition('=')
                extras[key] = value
            params.update(extras)
        return params


class UpdateDevice(command.UpdateCommand):
    """Update attributes of a device."""
    resource = 'device'
    json_indent = 4
    log = logging.getLogger(__name__ + '.UpdateDevice')
    name_key = 'name'

    def get_parser(self, prog_name):
        parser = super(UpdateDevice, self).get_parser(prog_name)
        parser.add_argument(
            '--device-type',
            help='Choose from container, vm, or shell. '
                 'Default to container'
        )
        parser.add_argument(
            '--device-driver',
            help='The driver of the device'
        )
        parser.add_argument(
            '--extra', metavar='<key>=<value>',
            action='append',
            dest='extra_capabilities',
            default=[],
            help='Extra capabilities key/value pairs to update for the device'
        )
        return parser

    def args2body(self, parsed_args):
        params = {}
        parsed_args_dict = vars(parsed_args)
        for arg in ['name', 'device_type', 'device_driver']:
            if parsed_args_dict.get(arg):
                params[arg] = parsed_args_dict.get(arg)
        extras = {}
        if parsed_args.extra_capabilities:
            for capa in parsed_args.extra_capabilities:
                key, _sep, value = capa.partition('=')
                extras[key] = value
            params['values'] = extras
        return params


class UnsetAttributeDevice(UpdateDevice):
    log = logging.getLogger(__name__ + '.UnsetAttributeDevice')

    def get_parser(self, prog_name):
        parser = super(UpdateDevice, self).get_parser(prog_name)
        parser.add_argument(
            '--extra', metavar='<key>',
            action='append',
            dest='extra_capabilities',
            default=[],
            help='Extra capability keys which should be unset from the device.',
        )
        return parser

    def args2body(self, parsed_args):
        if parsed_args.extra_capabilities:
            return {
                'values': {
                    cap: None for cap in parsed_args.extra_capabilities
                }
            }
        else:
            return {}


class DeleteDevice(command.DeleteCommand):
    """Delete a device."""
    resource = 'device'
    log = logging.getLogger(__name__ + '.DeleteDevice')
    name_key = 'name'


class ShowDeviceAllocation(command.ShowAllocationCommand):
    """Show device allocation details."""
    resource = 'device'
    json_indent = 4
    log = logging.getLogger(__name__ + '.ShowDeviceAllocation')


class ListDeviceAllocations(command.ListAllocationCommand):
    """List device allocations."""
    resource = 'device'
    log = logging.getLogger(__name__ + '.ListDeviceAllocations')
    list_columns = ['resource_id', 'reservations']

    def get_parser(self, prog_name):
        parser = super(ListDeviceAllocations, self).get_parser(prog_name)
        parser.add_argument(
            '--sort-by', metavar="<device_column>",
            help='column name used to sort result',
            default='resource_id'
        )
        return parser


class ReallocateDevice(command.ReallocateCommand):
    """Reallocate device from current allocations."""
    resource = 'device'
    json_indent = 4
    log = logging.getLogger(__name__ + '.ReallocateDevice')
    name_key = 'name'

    def get_parser(self, prog_name):
        parser = super(ReallocateDevice, self).get_parser(prog_name)
        parser.add_argument(
            '--lease-id',
            help='Lease ID to reallocate device from.')
        parser.add_argument(
            '--reservation-id',
            help='Reservation ID to reallocate device from')
        return parser

    def args2body(self, parsed_args):
        params = {}

        if parsed_args.reservation_id:
            params['reservation_id'] = parsed_args.reservation_id
        elif parsed_args.lease_id:
            params['lease_id'] = parsed_args.lease_id

        return params


class ShowDeviceProperty(command.ShowPropertyCommand):
    """Show device property."""
    resource = 'device'
    json_indent = 4
    log = logging.getLogger(__name__ + '.ShowDeviceProperty')


class ListDeviceProperties(command.ListCommand):
    """List device properties."""
    resource = 'device'
    log = logging.getLogger(__name__ + '.ListDeviceProperties')
    list_columns = ['property', 'private', 'property_values']

    def args2body(self, parsed_args):
        params = {
            'detail': parsed_args.detail,
            'all': parsed_args.all,
        }
        if parsed_args.sort_by:
            if parsed_args.sort_by in self.list_columns:
                params['sort_by'] = parsed_args.sort_by
            else:
                msg = 'Invalid sort option %s' % parsed_args.sort_by
                raise exception.BlazarClientException(msg)

        return params

    def retrieve_list(self, parsed_args):
        """Retrieve a list of resources from Blazar server."""
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.list_properties(**body)
        return data

    def get_parser(self, prog_name):
        parser = super(ListDeviceProperties, self).get_parser(prog_name)
        parser.add_argument(
            '--detail',
            action='store_true',
            help='Return properties with values and attributes.',
            default=False
        )
        parser.add_argument(
            '--sort-by', metavar="<property_column>",
            help='column name used to sort result',
            default='property'
        )
        parser.add_argument(
            '--all',
            action='store_true',
            help='Return all properties, public and private.',
            default=False
        )
        return parser


class UpdateDeviceProperty(command.UpdatePropertyCommand):
    """Update attributes of a device property."""
    resource = 'device'
    json_indent = 4
    log = logging.getLogger(__name__ + '.UpdateDeviceProperty')
    name_key = 'property_name'



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import re
import sys

import fixtures
#note(n.s.): you may need it later
#import mock
import testtools

#note(n.s.): you may need it later
#from blazarclient import client as blazar_client
#from blazarclient import exception
from blazarclient import shell
from blazarclient import tests

FAKE_ENV = {'OS_USERNAME': 'username',
            'OS_USER_DOMAIN_ID': 'user_domain_id',
            'OS_PASSWORD': 'password',
            'OS_TENANT_NAME': 'tenant_name',
            'OS_PROJECT_NAME': 'project_name',
            'OS_PROJECT_DOMAIN_ID': 'project_domain_id',
            'OS_AUTH_URL': 'http://no.where'}


class BlazarShellTestCase(tests.TestCase):

    def make_env(self, exclude=None, fake_env=FAKE_ENV):
        env = dict((k, v) for k, v in fake_env.items() if k != exclude)
        self.useFixture(fixtures.MonkeyPatch('os.environ', env))

    def setUp(self):
        super(BlazarShellTestCase, self).setUp()

        #Create shell for non-specific tests
        self.blazar_shell = shell.BlazarShell()

    def shell(self, argstr, exitcodes=(0,)):
        orig = sys.stdout
        orig_stderr = sys.stderr
        try:
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()
            _shell = shell.BlazarShell()
            _shell.initialize_app(argstr.split())
        except SystemExit:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.assertIn(exc_value.code, exitcodes)
        finally:
            stdout = sys.stdout.getvalue()
            sys.stdout.close()
            sys.stdout = orig
            stderr = sys.stderr.getvalue()
            sys.stderr.close()
            sys.stderr = orig_stderr
        return (stdout, stderr)

    def test_help_unknown_command(self):
        self.assertRaises(ValueError, self.shell, 'bash-completion')

    @testtools.skip('lol')
    def test_bash_completion(self):
        stdout, stderr = self.shell('bash-completion')
        # just check we have some output
        required = [
            '.*--matching',
            '.*--wrap',
            '.*help',
            '.*secgroup-delete-rule',
            '.*--priority']
        for r in required:
            self.assertThat((stdout + stderr),
                            testtools.matchers.MatchesRegex(
                                r, re.DOTALL | re.MULTILINE))

    @testtools.skip('lol')
    def test_authenticate_user(self):
        obj = shell.BlazarShell()
        obj.initialize_app('list-leases')
        obj.options.os_token = 'aaaa-bbbb-cccc'
        obj.options.os_cacert = 'cert'

        obj.authenticate_user()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from blazarclient.v1 import devices
from blazarclient.v1 import allocations
from blazarclient.v1 import floatingips
from blazarclient.v1 import hosts
from blazarclient.v1 import leases
from blazarclient.v1 import networks


class Client(object):
    """Top level object to communicate with Blazar.

    Contains managers to control requests that should be passed to each type of
    resources - leases, events, etc.

    **Examples**
        client = Client()
        client.lease.list()
        client.event.list(<lease_id>)
        ...
    """

    version = '1'

    def __init__(self, blazar_url=None, auth_token=None, session=None,
                 **kwargs):
        self.blazar_url = blazar_url
        self.auth_token = auth_token
        self.session = session

        if not self.session:
            logging.warning('Use a keystoneauth session object for the '
                            'authentication. The authentication with '
                            'blazar_url and auth_token is deprecated.')

        self.lease = leases.LeaseClientManager(blazar_url=self.blazar_url,
                                               auth_token=self.auth_token,
                                               session=self.session,
                                               version=self.version,
                                               **kwargs)
        self.host = hosts.ComputeHostClientManager(blazar_url=self.blazar_url,
                                                   auth_token=self.auth_token,
                                                   session=self.session,
                                                   version=self.version,
                                                   **kwargs)
        self.floatingip = floatingips.FloatingIPClientManager(
            blazar_url=self.blazar_url,
            auth_token=self.auth_token,
            session=self.session,
            version=self.version,
            **kwargs)
        self.network = networks.NetworkClientManager(
            blazar_url=self.blazar_url,
            auth_token=self.auth_token,
            session=self.session,
            version=self.version,
            **kwargs)
        self.device = devices.DeviceClientManager(
            blazar_url=self.blazar_url,
            auth_token=self.auth_token,
            session=self.session,
            version=self.version,
            **kwargs)
        self.allocation = allocations.AllocationClientManager(
            blazar_url=self.blazar_url,
            auth_token=self.auth_token,
            session=self.session,
            version=self.version,
            **kwargs)



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import mock

from blazarclient import base
from blazarclient import exception
from blazarclient import tests


class RequestManagerTestCase(tests.TestCase):

    def setUp(self):
        super(RequestManagerTestCase, self).setUp()

        self.blazar_url = "www.fake.com/reservation"
        self.auth_token = "aaa-bbb-ccc"
        self.user_agent = "python-blazarclient"
        self.manager = base.RequestManager(blazar_url=self.blazar_url,
                                           auth_token=self.auth_token,
                                           user_agent=self.user_agent)

    @mock.patch('blazarclient.base.RequestManager.request',
                return_value=(200, {"fake": "FAKE"}))
    def test_get(self, m):
        url = '/leases'
        resp, body = self.manager.get(url)
        self.assertEqual(resp, 200)
        self.assertDictEqual(body, {"fake": "FAKE"})
        m.assert_called_once_with(url, "GET")

    @mock.patch('blazarclient.base.RequestManager.request',
                return_value=(200, {"fake": "FAKE"}))
    def test_post(self, m):
        url = '/leases'
        req_body = {
            'start': '2020-07-24 20:00',
            'end': '2020-08-09 22:30',
            'before_end': '2020-08-09 21:30',
            'events': [],
            'name': 'lease-test',
            'reservations': [
                {
                    'min': '1',
                    'max': '2',
                    'hypervisor_properties':
                        '[">=", "$vcpus", "2"]',
                    'resource_properties':
                        '["==", "$extra_key", "extra_value"]',
                    'resource_type': 'physical:host',
                    'before_end': 'default'
                }
            ]
        }
        resp, body = self.manager.post(url, req_body)
        self.assertEqual(resp, 200)
        self.assertDictEqual(body, {"fake": "FAKE"})
        m.assert_called_once_with(url, "POST", body=req_body)

    @mock.patch('blazarclient.base.RequestManager.request',
                return_value=(200, {"fake": "FAKE"}))
    def test_delete(self, m):
        url = '/leases/aaa-bbb-ccc'
        resp, body = self.manager.delete(url)
        self.assertEqual(resp, 200)
        self.assertDictEqual(body, {"fake": "FAKE"})
        m.assert_called_once_with(url, "DELETE")

    @mock.patch('blazarclient.base.RequestManager.request',
                return_value=(200, {"fake": "FAKE"}))
    def test_put(self, m):
        url = '/leases/aaa-bbb-ccc'
        req_body = {
            'name': 'lease-test',
        }
        resp, body = self.manager.put(url, req_body)
        self.assertEqual(resp, 200)
        self.assertDictEqual(body, {"fake": "FAKE"})
        m.assert_called_once_with(url, "PUT", body=req_body)

    @mock.patch('requests.request')
    def test_request_ok_with_body(self, m):
        m.return_value.status_code = 200
        m.return_value.text = '{"resp_key": "resp_value"}'
        url = '/leases'
        kwargs = {"body": {"req_key": "req_value"}}
        self.assertEqual(self.manager.request(url, "POST", **kwargs),
                         (m(), {"resp_key": "resp_value"}))

    @mock.patch('requests.request')
    def test_request_ok_without_body(self, m):
        m.return_value.status_code = 200
        m.return_value.text = "resp"
        url = '/leases'
        kwargs = {"body": {"req_key": "req_value"}}
        self.assertEqual(self.manager.request(url, "POST", **kwargs),
                         (m(), None))

    @mock.patch('requests.request')
    def test_request_fail_with_body(self, m):
        m.return_value.status_code = 400
        m.return_value.text = '{"resp_key": "resp_value"}'
        url = '/leases'
        kwargs = {"body": {"req_key": "req_value"}}
        self.assertRaises(exception.BlazarClientException,
                          self.manager.request, url, "POST", **kwargs)

    @mock.patch('requests.request')
    def test_request_fail_without_body(self, m):
        m.return_value.status_code = 400
        m.return_value.text = "resp"
        url = '/leases'
        kwargs = {"body": {"req_key": "req_value"}}
        self.assertRaises(exception.BlazarClientException,
                          self.manager.request, url, "POST", **kwargs)


class SessionClientTestCase(tests.TestCase):

    def setUp(self):
        super(SessionClientTestCase, self).setUp()
        self.manager = base.SessionClient(user_agent="python-blazarclient",
                                          session=mock.MagicMock())

    @mock.patch('blazarclient.base.adapter.LegacyJsonAdapter.request')
    def test_request_ok(self, m):
        mock_resp = mock.Mock()
        mock_resp.status_code = 200
        mock_body = {"resp_key": "resp_value"}
        m.return_value = (mock_resp, mock_body)
        url = '/leases'
        kwargs = {"body": {"req_key": "req_value"}}
        resp, body = self.manager.request(url, "POST", **kwargs)
        self.assertEqual((resp, body), (mock_resp, mock_body))

    @mock.patch('blazarclient.base.adapter.LegacyJsonAdapter.request')
    def test_request_fail(self, m):
        resp = mock.Mock()
        resp.status_code = 400
        body = {"error message": "error"}
        m.return_value = (resp, body)
        url = '/leases'
        kwargs = {"body": {"req_key": "req_value"}}
        self.assertRaises(exception.BlazarClientException,
                          self.manager.request, url, "POST", **kwargs)


class BaseClientManagerTestCase(tests.TestCase):

    def setUp(self):
        super(BaseClientManagerTestCase, self).setUp()

        self.blazar_url = "www.fake.com/reservation"
        self.auth_token = "aaa-bbb-ccc"
        self.session = mock.MagicMock()
        self.user_agent = "python-blazarclient"

    def test_init_with_session(self):
        manager = base.BaseClientManager(blazar_url=None,
                                         auth_token=None,
                                         session=self.session)
        self.assertIsInstance(manager.request_manager,
                              base.SessionClient)

    def test_init_with_url_and_token(self):
        manager = base.BaseClientManager(blazar_url=self.blazar_url,
                                         auth_token=self.auth_token,
                                         session=None)
        self.assertIsInstance(manager.request_manager,
                              base.RequestManager)

    def test_init_with_insufficient_info(self):
        self.assertRaises(exception.InsufficientAuthInformation,
                          base.BaseClientManager,
                          blazar_url=None,
                          auth_token=self.auth_token,
                          session=None)



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazarclient import base
from blazarclient.i18n import _


class NetworkClientManager(base.BaseClientManager):
    """Manager for network segment requests."""

    def create(self, network_type, physical_network, segment_id, **kwargs):
        """Creates a network segment from values passed."""
        values = {'network_type': network_type,
                  'physical_network': physical_network,
                  'segment_id': segment_id}
        values.update(**kwargs)
        resp, body = self.request_manager.post('/networks', body=values)
        return body['network']

    def get(self, network_id):
        """Show network segment details."""
        resp, body = self.request_manager.get('/networks/%s' % network_id)
        return body['network']

    def update(self, network_id, values):
        """Update attributes of the network segment."""
        if not values:
            return _('No values to update passed.')
        resp, body = self.request_manager.put(
            '/networks/%s' % network_id, body=values
        )
        return body['network']

    def delete(self, network_id):
        """Delete network segment with specified ID."""
        resp, body = self.request_manager.delete('/networks/%s' % network_id)

    def list(self, sort_by=None):
        """List all network segments."""
        resp, body = self.request_manager.get('/networks')
        networks = body['networks']
        if sort_by:
            networks = sorted(networks, key=lambda network: network[sort_by])
        return networks

    def get_allocation(self, network_id):
        """Get allocation for network."""
        resp, body = self.request_manager.get(
            '/networks/%s/allocation' % network_id)
        return body['allocation']

    def list_allocations(self, sort_by=None):
        """List allocations for all networks."""
        resp, body = self.request_manager.get('/networks/allocations')
        allocations = body['allocations']
        if sort_by:
            allocations = sorted(allocations, key=lambda alloc: alloc[sort_by])
        return allocations

    def list_properties(self, detail=False, all=False, sort_by=None):
        url = '/networks/properties'

        if detail:
            url += '?detail=True'

        resp, body = self.request_manager.get(url)
        resource_properties = body['resource_properties']

        # Values is a reserved word in cliff so need to rename values column.
        if detail:
            for p in resource_properties:
                p['property_values'] = p['values']
                del p['values']

        if sort_by:
            resource_properties = sorted(resource_properties,
                                         key=lambda prop: prop[sort_by])
        return resource_properties

    def get_property(self, property_name):
        resource_property = [
            x for x in self.list_properties(detail=True)
            if x['property'] == property_name]

        return {} if not resource_property else resource_property[0]

    def set_property(self, property_name, private):
        data = {'private': private}
        resp, body = self.request_manager.patch(
            '/networks/properties/%s' % property_name, body=data)

        return body['resource_property']



# Copyright (c) 2018 StackHPC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from unittest import mock

from blazarclient import shell
from blazarclient import tests
from blazarclient.v1.networks import NetworkClientManager
from blazarclient.v1.shell_commands import networks


class CreateNetworkTest(tests.TestCase):

    def setUp(self):
        super(CreateNetworkTest, self).setUp()
        self.create_network = networks.CreateNetwork(shell.BlazarShell(), mock.Mock())

    def test_args2body(self):
        args = argparse.Namespace(
            network_type='vlan',
            physical_network='physnet1',
            segment_id='1234',
            extra_capabilities=[
                'extra_key1=extra_value1',
                'extra_key2=extra_value2',
            ]
        )

        expected = {
            'network_type': 'vlan',
            'physical_network': 'physnet1',
            'segment_id': '1234',
            'extra_key1': 'extra_value1',
            'extra_key2': 'extra_value2',
        }

        ret = self.create_network.args2body(args)
        self.assertDictEqual(ret, expected)


class UpdateNetworkTest(tests.TestCase):

    def create_update_command(self, list_value):
        mock_network_manager = mock.Mock()
        mock_network_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.network = mock_network_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return networks.UpdateNetwork(blazar_shell, mock.Mock()), mock_network_manager

    def test_update_network(self):
        list_value = [
            {'id': '072c58c0-64ac-467b-b040-9138771e146a', 'networkname': 'network-1'},
            {'id': '072c58c0-64ac-467b-b040-9138771e146a', 'networkname': 'network-2'},
        ]
        update_network, network_manager = self.create_update_command(list_value)
        args = argparse.Namespace(
            id='072c58c0-64ac-467b-b040-9138771e146a',
            extra_capabilities=[
                'key1=value1',
                'key2=value2'
            ])
        expected = {
            'values': {
                'key1': 'value1',
                'key2': 'value2'
            }
        }
        update_network.run(args)
        network_manager.update.assert_called_once_with('072c58c0-64ac-467b-b040-9138771e146a', **expected)


class UnsetAttributesNetworkTest(tests.TestCase):

    def create_unset_command(self, list_value):
        mock_network_manager = mock.Mock()
        mock_network_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.network = mock_network_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return networks.UnsetAttributeNetwork(
            blazar_shell, mock.Mock()
        ), mock_network_manager

    def test_unset_network(self):
        list_value = [
            {'id': '072c58c0-64ac-467b-b040-9138771e146a', 'networkname': 'network-1'},
            {'id': '072c58c0-64ac-467b-b040-9138771e146b', 'networkname': 'network-2'},
        ]
        unset_network, network_manager = self.create_unset_command(list_value)
        extra_caps = ['key1', 'key2']
        args = argparse.Namespace(
            id='072c58c0-64ac-467b-b040-9138771e146a',
            extra_capabilities=extra_caps,
        )
        expected = {
            'values': {key: None for key in extra_caps}
       }
        unset_network.run(args)
        network_manager.update.assert_called_once_with('072c58c0-64ac-467b-b040-9138771e146a', **expected)

class ShowNetworkTest(tests.TestCase):

    def create_show_command(self, list_value, get_value):
        mock_network_manager = mock.Mock()
        mock_network_manager.list.return_value = list_value
        mock_network_manager.get.return_value = get_value

        mock_client = mock.Mock()
        mock_client.network = mock_network_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return networks.ShowNetwork(blazar_shell, mock.Mock()), mock_network_manager

    def test_show_network(self):
        list_value = [
            {'id': '072c58c0-64ac-467b-b040-9138771e146a'},
            {'id': '072c58c0-64ac-467b-b040-9138771e146b'},
        ]
        get_value = {
            'id': '072c58c0-64ac-467b-b040-9138771e146a'}

        show_network, network_manager = self.create_show_command(list_value,
                                                           get_value)

        args = argparse.Namespace(id='072c58c0-64ac-467b-b040-9138771e146a', formatter="table")
        expected = [('id',), ('072c58c0-64ac-467b-b040-9138771e146a',)]

        ret = show_network.get_data(args)
        self.assertEqual(ret, expected)

        network_manager.get.assert_called_once_with('072c58c0-64ac-467b-b040-9138771e146a')


class DeleteNetworkTest(tests.TestCase):

    def create_delete_command(self, list_value):
        mock_network_manager = mock.Mock()
        mock_network_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.network = mock_network_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return networks.DeleteNetwork(blazar_shell, mock.Mock()), mock_network_manager

    def test_delete_network(self):
        # ID should be of format of blazarclient.command.UUID_PATTERN
        list_value = [
            {'id': '072c58c0-64ac-467b-b040-9138771e146a', 'networkname': 'network-1'},
            {'id': '072c58c0-64ac-467b-b040-9138771e146b', 'networkname': 'network-2'},
        ]
        delete_network, network_manager = self.create_delete_command(list_value)

        args = argparse.Namespace(id='072c58c0-64ac-467b-b040-9138771e146a')
        delete_network.run(args)

        network_manager.delete.assert_called_once_with('072c58c0-64ac-467b-b040-9138771e146a')

class MockNetworkClientManager(NetworkClientManager):
    def __init__(self):
        self.request_manager = MockRequestManager()

class MockRequestManager:
    def get(self, url):
        return "-", {
            "networks":
            [
                {
                    "id": "256c0f35-b29e-45cb-9931-ea785f415955",
                    "network_type": "vlan",
                    "physical_network": "physnet1",
                    "segment_id": 3008
                },
                {
                    "id": "937ad993-53c6-4055-9802-e8f360f46598",
                    "network_type": "vlan",
                    "physical_network": "physnet1",
                    "segment_id": 3294
                },
                {
                    "id": "d97e5bb4-46e2-4b65-a460-a2d0bb305d29",
                    "network_type": "vlan",
                    "physical_network": "physnet1",
                    "segment_id": 3006
                },
                {
                    "id": "ee5ec1b8-7295-4c56-8f85-4c89317785d9",
                    "network_type": "vlan",
                    "physical_network": "physnet1",
                    "segment_id": 3009
                }
            ]
        }


class ListNetworksTest(tests.TestCase):

    def create_list_command(self):
        mock_network_manager = MockNetworkClientManager()

        mock_client = mock.Mock()
        mock_client.network = mock_network_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return networks.ListNetworks(blazar_shell, mock.Mock()), mock_network_manager

    def test_list_network_sort_by(self):

        list_network, network_manager = self.create_list_command()
        list_networks_args = argparse.Namespace(
            sort_by='segment_id',
            columns=[],
            formatter="table",
            long=False,
        )
        return_networks = list_network.get_data(list_networks_args)
        segment_id_index = list(return_networks[0]).index('segment_id')
        prev_segment_id = 0
        for network in list(return_networks[1]):
            network_segment_id = network[segment_id_index]
            self.assertLess(prev_segment_id, network_segment_id)
            prev_segment_id = network_segment_id
        # network_manager.list.assert_called_once_with(sort_by='segment_id')



# Copyright (c) 2019 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from blazarclient import command


class ListFloatingIPs(command.ListCommand):
    """Print a list of floating IPs."""
    resource = 'floatingip'
    log = logging.getLogger(__name__ + '.ListFloatingIPs')
    list_columns = ['id', 'floating_ip_address', 'floating_network_id']

    def get_parser(self, prog_name):
        parser = super(ListFloatingIPs, self).get_parser(prog_name)
        parser.add_argument(
            '--sort-by', metavar="<floatingip_column>",
            help='column name used to sort result',
            default='id'
        )
        return parser


class ShowFloatingIP(command.ShowCommand):
    """Show floating IP details."""
    resource = 'floatingip'
    allow_names = False
    json_indent = 4
    log = logging.getLogger(__name__ + '.ShowFloatingIP')

    def get_parser(self, prog_name):
        parser = super(ShowFloatingIP, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'ID or name of %s to look up'
        else:
            help_str = 'ID of %s to look up'
        parser.add_argument('id', metavar=self.resource.upper(),
                            help=help_str % self.resource)
        return parser


class CreateFloatingIP(command.CreateCommand):
    """Create a floating IP."""
    resource = 'floatingip'
    json_indent = 4
    log = logging.getLogger(__name__ + '.CreateFloatingIP')

    def get_parser(self, prog_name):
        parser = super(CreateFloatingIP, self).get_parser(prog_name)
        parser.add_argument(
            'network_id', metavar='NETWORK_ID',
            help='External network ID to which the floating IP belongs'
        )
        parser.add_argument(
            'floating_ip_address', metavar='FLOATING_IP_ADDRESS',
            help='Floating IP address to add to Blazar'
        )
        return parser

    def args2body(self, parsed_args):
        params = {}
        if parsed_args.network_id:
            params['network_id'] = parsed_args.network_id
        if parsed_args.floating_ip_address:
            params['floating_ip_address'] = parsed_args.floating_ip_address
        return params


class DeleteFloatingIP(command.DeleteCommand):
    """Delete a floating IP."""
    resource = 'floatingip'
    allow_names = False
    log = logging.getLogger(__name__ + '.DeleteFloatingIP')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from blazarclient.i18n import _


class BlazarClientException(Exception):
    """Base exception class."""
    message = _("An unknown exception occurred %s.")
    code = 500

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            message = self.message % kwargs

        super(BlazarClientException, self).__init__(message)


class CommandError(BlazarClientException):
    """Occurs if not all authentication vital options are set."""
    message = _("You have to provide all options like user name or tenant "
                "id to make authentication possible.")
    code = 401


class NotAuthorized(BlazarClientException):
    """HTTP 401 - Not authorized.

    User have no enough rights to perform action.
    """
    code = 401
    message = _("Not authorized request.")


class NoBlazarEndpoint(BlazarClientException):
    """Occurs if no endpoint for Blazar set in the Keystone."""
    message = _("No publicURL endpoint for Blazar found. Set endpoint "
                "for Blazar in the Keystone.")
    code = 404


class NoUniqueMatch(BlazarClientException):
    """Occurs if there are more than one appropriate resources."""
    message = _("There is no unique requested resource.")
    code = 409


class UnsupportedVersion(BlazarClientException):
    """Occurs if unsupported client version was requested."""
    message = _("Unsupported client version requested.")
    code = 406


class IncorrectLease(BlazarClientException):
    """Occurs if lease parameters are incorrect."""
    message = _("The lease parameters are incorrect.")
    code = 409


class DuplicatedLeaseParameters(BlazarClientException):
    """Occurs if lease parameters are duplicated."""
    message = _("The lease parameters are duplicated.")
    code = 400


class InsufficientAuthInformation(BlazarClientException):
    """Occurs if the auth info passed to blazar client is insufficient."""
    message = _("The passed arguments are insufficient "
                "for the authentication. The instance of "
                "keystoneauth1.session.Session class is required.")
    code = 400


class IncorrectNetwork(BlazarClientException):
    """Occurs if network parameters are incorrect."""
    message = _("The network parameters are incorrect.")
    code = 409


class ResourcePropertyNotFound(BlazarClientException):
    """Occurs if the resource property specified does not exist"""
    message = _("The resource property does not exist.")
    code = 404



# Copyright (c) 2017 NTT Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from datetime import datetime
from unittest import mock

from blazarclient import exception
from blazarclient import shell
from blazarclient import tests
from blazarclient.v1.shell_commands import leases

mock_time = mock.Mock(return_value=datetime(2020, 6, 8))

FIRST_LEASE = 'd1e43d6d-8f6f-4c2e-b0a9-2982b39dc698'
SECOND_LEASE = '424d21c3-45a2-448a-81ad-32eddc888375'


@mock.patch('blazarclient.v1.shell_commands.leases._utc_now', mock_time)
class CreateLeaseTestCase(tests.TestCase):

    def setUp(self):
        super(CreateLeaseTestCase, self).setUp()
        self.cl = leases.CreateLease(shell.BlazarShell(), mock.Mock())

    def test_args2body_correct_phys_res_params(self):
        args = argparse.Namespace(
            start='2020-07-24 20:00',
            end='2020-08-09 22:30',
            before_end='2020-08-09 21:30',
            events=[],
            name='lease-test',
            reservations=[],
            physical_reservations=[
                'min=1,'
                'max=2,'
                'hypervisor_properties='
                '["and", [">=", "$vcpus", "2"], '
                '[">=", "$memory_mb", "2048"]],'
                'resource_properties='
                '["==", "$extra_key", "extra_value"],'
                'before_end=default'
            ]
        )
        expected = {
            'start': '2020-07-24 20:00',
            'end': '2020-08-09 22:30',
            'before_end': '2020-08-09 21:30',
            'events': [],
            'name': 'lease-test',
            'reservations': [
                {
                    'min': 1,
                    'max': 2,
                    'hypervisor_properties':
                        '["and", [">=", "$vcpus", "2"], '
                        '[">=", "$memory_mb", "2048"]]',
                    'resource_properties':
                        '["==", "$extra_key", "extra_value"]',
                    'resource_type': 'physical:host',
                    'before_end': 'default'
                }
            ]
        }
        self.assertDictEqual(self.cl.args2body(args), expected)

    def test_args2body_incorrect_phys_res_params(self):
        args = argparse.Namespace(
            start='2020-07-24 20:00',
            end='2020-08-09 22:30',
            before_end='2020-08-09 21:30',
            events=[],
            name='lease-test',
            reservations=[],
            physical_reservations=[
                'incorrect_param=1,'
                'min=1,'
                'max=2,'
                'hypervisor_properties='
                '["and", [">=", "$vcpus", "2"], '
                '[">=", "$memory_mb", "2048"]],'
                'resource_properties='
                '["==", "$extra_key", "extra_value"]'
            ]
        )
        self.assertRaises(exception.IncorrectLease,
                          self.cl.args2body,
                          args)

    def test_args2body_duplicated_phys_res_params(self):
        args = argparse.Namespace(
            start='2020-07-24 20:00',
            end='2020-08-09 22:30',
            before_end='2020-08-09 21:30',
            events=[],
            name='lease-test',
            reservations=[],
            physical_reservations=[
                'min=1,'
                'min=1,'
                'max=2,'
                'hypervisor_properties='
                '["and", [">=", "$vcpus", "2"], '
                '[">=", "$memory_mb", "2048"]],'
                'resource_properties='
                '["==", "$extra_key", "extra_value"]'
            ]
        )
        self.assertRaises(exception.DuplicatedLeaseParameters,
                          self.cl.args2body,
                          args)

    def test_args2body_correct_instance_res_params(self):
        args = argparse.Namespace(
            start='2020-07-24 20:00',
            end='2020-08-09 22:30',
            before_end='2020-08-09 21:30',
            events=[],
            name='lease-test',
            reservations=[
                'vcpus=4,'
                'memory_mb=1024,'
                'disk_gb=10,'
                'amount=2,'
                'affinity=True,'
                'resource_properties='
                '["==", "$extra_key", "extra_value"],'
                'resource_type=virtual:instance'
            ],
            physical_reservations=[
                'min=1,'
                'max=2,'
                'hypervisor_properties='
                '["and", [">=", "$vcpus", "2"], '
                '[">=", "$memory_mb", "2048"]],'
                'resource_properties='
                '["==", "$extra_key", "extra_value"],'
                'before_end=default'
            ]
        )
        expected = {
            'start': '2020-07-24 20:00',
            'end': '2020-08-09 22:30',
            'before_end': '2020-08-09 21:30',
            'events': [],
            'name': 'lease-test',
            'reservations': [
                {
                    'min': 1,
                    'max': 2,
                    'hypervisor_properties':
                        '["and", [">=", "$vcpus", "2"], '
                        '[">=", "$memory_mb", "2048"]]',
                    'resource_properties':
                        '["==", "$extra_key", "extra_value"]',
                    'resource_type': 'physical:host',
                    'before_end': 'default'
                },
                {
                    'vcpus': 4,
                    'memory_mb': 1024,
                    'disk_gb': 10,
                    'amount': 2,
                    'affinity': 'True',
                    'resource_properties':
                        '["==", "$extra_key", "extra_value"]',
                    'resource_type': 'virtual:instance'
                }
            ]
        }
        self.assertDictEqual(self.cl.args2body(args), expected)

    def test_args2body_start_now(self):
        args = argparse.Namespace(
            start='now',
            end='2030-08-09 22:30',
            before_end='2030-08-09 21:30',
            events=[],
            name='lease-test',
            reservations=[],
            physical_reservations=[
                'min=1,'
                'max=2,'
                'hypervisor_properties='
                '["and", [">=", "$vcpus", "2"], '
                '[">=", "$memory_mb", "2048"]],'
                'resource_properties='
                '["==", "$extra_key", "extra_value"],'
                'before_end=default'
            ]
        )
        expected = {
            'start': 'now',
            'end': '2030-08-09 22:30',
            'before_end': '2030-08-09 21:30',
            'events': [],
            'name': 'lease-test',
            'reservations': [
                {
                    'min': 1,
                    'max': 2,
                    'hypervisor_properties':
                        '["and", [">=", "$vcpus", "2"], '
                        '[">=", "$memory_mb", "2048"]]',
                    'resource_properties':
                        '["==", "$extra_key", "extra_value"]',
                    'resource_type': 'physical:host',
                    'before_end': 'default'
                }
            ]
        }
        self.assertDictEqual(self.cl.args2body(args), expected)


class UpdateLeaseTestCase(tests.TestCase):

    def setUp(self):
        super(UpdateLeaseTestCase, self).setUp()
        self.cl = leases.UpdateLease(shell.BlazarShell(), mock.Mock())

    def test_args2body_time_params(self):
        args = argparse.Namespace(
            name=None,
            prolong_for='1h',
            reduce_by=None,
            end_date=None,
            defer_by=None,
            advance_by=None,
            start_date=None,
            reservation=None
        )
        expected = {
            'prolong_for': '1h',
        }

        self.assertDictEqual(self.cl.args2body(args), expected)

    def test_args2body_host_reservation_params(self):
        args = argparse.Namespace(
            name=None,
            prolong_for=None,
            reduce_by=None,
            end_date=None,
            defer_by=None,
            advance_by=None,
            start_date=None,
            reservation=[
                'id=798379a6-194c-45dc-ba34-1b5171d5552f,'
                'max=3,'
                'hypervisor_properties='
                '["and", [">=", "$vcpus", "4"], '
                '[">=", "$memory_mb", "8192"]],'
                'resource_properties='
                '["==", "$extra_key", "extra_value"]'
            ]
        )
        expected = {
            'reservations': [
                {
                    'id': '798379a6-194c-45dc-ba34-1b5171d5552f',
                    'max': 3,
                    'hypervisor_properties':
                        '["and", [">=", "$vcpus", "4"], '
                        '[">=", "$memory_mb", "8192"]]',
                    'resource_properties':
                        '["==", "$extra_key", "extra_value"]'
                }
            ]
        }

        self.assertDictEqual(self.cl.args2body(args), expected)

    def test_args2body_instance_reservation_params(self):
        args = argparse.Namespace(
            name=None,
            prolong_for=None,
            reduce_by=None,
            end_date=None,
            defer_by=None,
            advance_by=None,
            start_date=None,
            reservation=[
                'id=798379a6-194c-45dc-ba34-1b5171d5552f,'
                'vcpus=3,memory_mb=1024,disk_gb=20,'
                'amount=4,affinity=False'
            ]
        )
        expected = {
            'reservations': [
                {
                    'id': '798379a6-194c-45dc-ba34-1b5171d5552f',
                    'vcpus': 3,
                    'memory_mb': 1024,
                    'disk_gb': 20,
                    'amount': 4,
                    'affinity': 'False'
                }
            ]
        }

        self.assertDictEqual(self.cl.args2body(args), expected)


class ShowLeaseTestCase(tests.TestCase):

    def create_show_command(self):
        mock_lease_manager = mock.Mock()
        mock_client = mock.Mock()
        mock_client.lease = mock_lease_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return (leases.ShowLease(blazar_shell, mock.Mock()),
                mock_lease_manager)

    def test_show_lease(self):
        show_lease, lease_manager = self.create_show_command()
        lease_manager.get.return_value = {'id': FIRST_LEASE}
        lease_manager.list.return_value = [
            {'id': FIRST_LEASE, 'name': 'first-lease'},
            {'id': SECOND_LEASE, 'name': 'second-lease'},
        ]
        mock.seal(lease_manager)

        args = argparse.Namespace(id=FIRST_LEASE, detail=False, formatter="table")
        expected = [('id',), (FIRST_LEASE,)]

        self.assertEqual(show_lease.get_data(args), expected)
        lease_manager.get.assert_called_once_with(FIRST_LEASE)

    def test_show_lease_with_allocations(self):
        show_lease, lease_manager = self.create_show_command()
        lease_manager.get.return_value = {'id': FIRST_LEASE}
        lease_manager.additional_details.return_value = {
            'hosts': [],
            'networks': [],
            'devices': []
        }
        lease_manager.list.return_value = [
            {'id': FIRST_LEASE, 'name': 'first-lease'},
            {'id': SECOND_LEASE, 'name': 'second-lease'},
        ]
        mock.seal(lease_manager)

        args = argparse.Namespace(id=FIRST_LEASE, detail=True, formatter="table")
        expected = [('devices', 'hosts', 'id', 'networks',), ('[]', '[]', FIRST_LEASE, '[]',)]

        self.assertEqual(show_lease.get_data(args), expected)
        lease_manager.get.assert_called_once_with(FIRST_LEASE)
        lease_manager.additional_details.assert_called_once_with(FIRST_LEASE)

    def test_show_lease_by_name(self):
        show_lease, lease_manager = self.create_show_command()
        lease_manager.list.return_value = [
            {'id': FIRST_LEASE, 'name': 'first-lease'},
            {'id': SECOND_LEASE, 'name': 'second-lease'},
        ]
        lease_manager.get.return_value = {'id': SECOND_LEASE}
        mock.seal(lease_manager)

        args = argparse.Namespace(id='second-lease', detail=False, formatter="table")
        expected = [('id',), (SECOND_LEASE,)]

        self.assertEqual(show_lease.get_data(args), expected)
        lease_manager.list.assert_called_once_with()
        lease_manager.get.assert_called_once_with(SECOND_LEASE)

    def test_show_lease_by_name_with_allocations(self):
        show_lease, lease_manager = self.create_show_command()
        lease_manager.list.return_value = [
            {'id': FIRST_LEASE, 'name': 'first-lease'},
            {'id': SECOND_LEASE, 'name': 'second-lease'},
        ]
        lease_manager.get.return_value = {'id': SECOND_LEASE}
        host1 = {'id': '101', 'hypervisor_hostname': 'host-1'}
        lease_manager.additional_details.return_value = {
            'hosts': [host1],
            'networks': [],
            'devices': []
        }
        mock.seal(lease_manager)
        args = argparse.Namespace(id='second-lease', detail=True, formatter="table")
        expected = [
            ('devices', 'hosts', 'id', 'networks',),
            ('[]', '[\n    {\n        "id": "101",\n        "hypervisor_hostname": "host-1"\n    }\n]', '424d21c3-45a2-448a-81ad-32eddc888375', '[]')
        ]
        self.assertEqual(show_lease.get_data(args), expected)
        lease_manager.list.assert_called_once_with()
        lease_manager.get.assert_called_once_with(SECOND_LEASE)
        lease_manager.additional_details.assert_called_once_with(SECOND_LEASE)


class DeleteLeaseTestCase(tests.TestCase):

    def create_delete_command(self):
        mock_lease_manager = mock.Mock()
        mock_client = mock.Mock()
        mock_client.lease = mock_lease_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return (leases.DeleteLease(blazar_shell, mock.Mock()),
                mock_lease_manager)

    def test_delete_lease(self):
        delete_lease, lease_manager = self.create_delete_command()
        lease_manager.delete.return_value = None
        lease_manager.list.return_value = [{'id': FIRST_LEASE}]
        mock.seal(lease_manager)

        args = argparse.Namespace(id=FIRST_LEASE)
        delete_lease.run(args)

        lease_manager.delete.assert_called_once_with(FIRST_LEASE)

    def test_delete_lease_by_name(self):
        delete_lease, lease_manager = self.create_delete_command()
        lease_manager.list.return_value = [
            {'id': FIRST_LEASE, 'name': 'first-lease'},
            {'id': SECOND_LEASE, 'name': 'second-lease'},
        ]
        lease_manager.delete.return_value = None
        mock.seal(lease_manager)

        args = argparse.Namespace(id='second-lease')
        delete_lease.run(args)

        lease_manager.list.assert_called_once_with()
        lease_manager.delete.assert_called_once_with(SECOND_LEASE)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1 import adapter
from oslo_serialization import jsonutils
import requests

from blazarclient import exception
from blazarclient.i18n import _


class RequestManager(object):
    """Manager to create request from given Blazar URL and auth token."""

    def __init__(self, blazar_url, auth_token, user_agent):
        self.blazar_url = blazar_url
        self.auth_token = auth_token
        self.user_agent = user_agent

    def get(self, url):
        """Sends get request to Blazar.

        :param url: URL to the wanted Blazar resource.
        :type url: str
        """
        return self.request(url, 'GET')

    def post(self, url, body):
        """Sends post request to Blazar.

        :param url: URL to the wanted Blazar resource.
        :type url: str

        :param body: Values resource to be created from.
        :type body: dict
        """
        return self.request(url, 'POST', body=body)

    def delete(self, url):
        """Sends delete request to Blazar.

        :param url: URL to the wanted Blazar resource.
        :type url: str
        """
        return self.request(url, 'DELETE')

    def put(self, url, body):
        """Sends update request to Blazar.

        :param url: URL to the wanted Blazar resource.
        :type url: str

        :param body: Values resource to be updated from.
        :type body: dict
        """
        return self.request(url, 'PUT', body=body)

    def patch(self, url, body):
        """Sends patch request to Blazar.

        :param url: URL to the wanted Blazar resource.
        :type url: str
        """
        return self.request(url, 'PATCH', body=body)

    def request(self, url, method, **kwargs):
        """Base request method.

        Adds specific headers and URL prefix to the request.

        :param url: Resource URL.
        :type url: str

        :param method: Method to be called (GET, POST, PUT, DELETE).
        :type method: str

        :returns: Response and body.
        :rtype: tuple
        """
        kwargs.setdefault('headers', kwargs.get('headers', {}))
        kwargs['headers']['User-Agent'] = self.user_agent
        kwargs['headers']['Accept'] = 'application/json'
        kwargs['headers']['x-auth-token'] = self.auth_token

        if 'body' in kwargs:
            kwargs['headers']['Content-Type'] = 'application/json'
            kwargs['data'] = jsonutils.dump_as_bytes(kwargs['body'])
            del kwargs['body']

        resp = requests.request(method, self.blazar_url + url, **kwargs)

        try:
            body = jsonutils.loads(resp.text)
        except ValueError:
            body = None

        if resp.status_code >= 400:
            if body is not None:
                error_message = body.get('error_message', body)
            else:
                error_message = resp.text

            body = _("ERROR: {0}").format(error_message)
            raise exception.BlazarClientException(body, code=resp.status_code)

        return resp, body


class SessionClient(adapter.LegacyJsonAdapter):
    """Manager to create request with keystoneauth1 session."""

    def request(self, url, method, **kwargs):
        resp, body = super(SessionClient, self).request(
            url, method, raise_exc=False, **kwargs)

        if resp.status_code >= 400:
            if body is not None:
                error_message = body.get('error_message', body)
            else:
                error_message = resp.text

            msg = _("ERROR: {0}").format(error_message)
            raise exception.BlazarClientException(msg, code=resp.status_code)

        return resp, body


class BaseClientManager(object):
    """Base class for managing resources of Blazar."""

    user_agent = 'python-blazarclient'

    def __init__(self, blazar_url, auth_token, session, **kwargs):
        self.blazar_url = blazar_url
        self.auth_token = auth_token
        self.session = session

        if self.session:
            self.request_manager = SessionClient(
                session=self.session,
                user_agent=self.user_agent,
                **kwargs
            )
        elif self.blazar_url and self.auth_token:
            self.request_manager = RequestManager(blazar_url=self.blazar_url,
                                                  auth_token=self.auth_token,
                                                  user_agent=self.user_agent)
        else:
            raise exception.InsufficientAuthInformation



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from blazarclient.osc import plugin
from blazarclient import tests


class ReservationPluginTests(tests.TestCase):

    @mock.patch("blazarclient.v1.client.Client")
    def test_make_client(self, mock_client):
        instance = mock.Mock()
        instance._api_version = {"reservation": "1"}
        endpoint = "blazar_endpoint"
        instance.get_endpoint_for_service_type = mock.Mock(
            return_value=endpoint
        )

        plugin.make_client(instance)

        mock_client.assert_called_with(
            "1",
            session=instance.session,
            endpoint_override=endpoint
        )



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from unittest import mock

import testtools

from blazarclient import command
from blazarclient import tests


class OpenstackCommandTestCase(tests.TestCase):

    def setUp(self):
        super(OpenstackCommandTestCase, self).setUp()

    @testtools.skip("Have no idea how to test super")
    def test_run(self):
        pass

    @testtools.skip("Unskip it when get_data will do smthg")
    def test_get_data(self):
        pass

    @testtools.skip("Unskip it when get_data will do smthg")
    def test_take_action(self):
        pass


class TableFormatterTestCase(tests.TestCase):

    def setUp(self):
        super(TableFormatterTestCase, self).setUp()

    @testtools.skip("Have no idea how to test super")
    def test_emit_list(self):
        pass


class BlazarCommandTestCase(tests.TestCase):

    def setUp(self):
        super(BlazarCommandTestCase, self).setUp()

        self.app = mock.MagicMock()
        self.parser = self.patch(command.OpenStackCommand, 'get_parser')

        self.command = command.BlazarCommand(self.app, [])

    def test_get_client(self):
        # Test that either client_manager.reservation or client is used,
        # whichever exists

        client_manager = self.app.client_manager
        del self.app.client_manager
        client = self.command.get_client()
        self.assertEqual(self.app.client, client)

        self.app.client_manager = client_manager
        del self.app.client
        client = self.command.get_client()
        self.assertEqual(self.app.client_manager.reservation, client)

    def test_get_parser(self):
        self.command.get_parser('TestCase')
        self.parser.assert_called_once_with('TestCase')

    def test_format_output_data(self):
        data_before = {'key_string': 'string_value',
                       'key_dict': {'key': 'value'},
                       'key_list': ['1', '2', '3'],
                       'key_none': None}
        data_after = {'key_string': 'string_value',
                      'key_dict': '{"key": "value"}',
                      'key_list': '["1", "2", "3"]',
                      'key_none': ''}

        args = argparse.Namespace(formatter="table")
        self.command.format_output_data(data_before, args)

        self.assertEqual(data_after, data_before)


class CreateCommandTestCase(tests.TestCase):
    def setUp(self):
        super(CreateCommandTestCase, self).setUp()

        self.app = mock.MagicMock()
        self.create_command = command.CreateCommand(self.app, [])

        self.client = self.patch(self.create_command, 'get_client')

    @testtools.skip("Under construction")
    def test_get_data_data(self):
        data = {'key_string': 'string_value',
                'key_dict': "{'key0': 'value', 'key1': 'value'}",
                'key_list': "['1', '2', '3',]",
                'key_none': None}
        self.client.resource.return_value = mock.MagicMock(return_value=data)
        self.assertEqual(self.create_command.get_data({'a': 'b'}), None)


@testtools.skip("Under construction")
class UpdateCommandTestCase(tests.TestCase):
    def setUp(self):
        super(UpdateCommandTestCase, self).setUp()

        self.app = mock.MagicMock()
        self.update_command = command.UpdateCommand(self.app, [])


@testtools.skip("Under construction")
class DeleteCommandTestCase(tests.TestCase):
    def setUp(self):
        super(DeleteCommandTestCase, self).setUp()

        self.app = mock.MagicMock()
        self.delete_command = command.DeleteCommand(self.app, [])


@testtools.skip("Under construction")
class ListCommandTestCase(tests.TestCase):
    def setUp(self):
        super(ListCommandTestCase, self).setUp()

        self.app = mock.MagicMock()
        self.list_command = command.ListCommand(self.app, [])


@testtools.skip("Under construction")
class ShowCommandTestCase(tests.TestCase):
    def setUp(self):
        super(ShowCommandTestCase, self).setUp()

        self.app = mock.MagicMock()
        self.show_command = command.ShowCommand(self.app, [])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re

from oslo_serialization import jsonutils
from oslo_utils import strutils

import argparse
from blazarclient import command
from blazarclient import exception
import logging


# All valid reservation parameters must be added to CREATE_RESERVATION_KEYS to
# make them parsable. Note that setting the default value to None ensures that
# the parameter is not included in the POST request if absent.
CREATE_RESERVATION_KEYS = {
    "network": {
        "network_name": "",
        "network_description": "",
        "network_properties": "",
        "resource_properties": "",
        "resource_type": 'network'
    },
    "physical:host": {
        "min": "",
        "max": "",
        "hypervisor_properties": "",
        "resource_properties": "",
        "before_end": None,
        "on_start": None,
        "resource_type": 'physical:host'
    },
    "virtual:floatingip": {
        "amount": 1,
        "network_id": None,
        "required_floatingips": [],
        "resource_type": 'virtual:floatingip'
    },
    "virtual:instance": {
        "vcpus": "",
        "memory_mb": "",
        "disk_gb": "",
        "amount": "",
        "affinity": "None",
        "resource_properties": "",
        "resource_type": 'virtual:instance'
    },
    "device": {
        "min": "",
        "max": "",
        "resource_properties": "",
        "resource_type": 'device'
    },
    "others": {
        ".*": None
    }
}


def _utc_now():
    """Wrap datetime.datetime.utcnow so it can be mocked in unit tests.

    This is required because some of the tests require understanding the
    'current time'; simply mocking utcnow() is made very difficult by
    the many different ways the datetime package is used in this module.
    """
    return datetime.datetime.utcnow()


class ListLeases(command.ListCommand):
    """Print a list of leases."""
    resource = 'lease'
    log = logging.getLogger(__name__ + '.ListLeases')
    list_columns = ['id', 'name', 'start_date', 'end_date']
    long_columns = ["status", "created_at", "degraded"]

    def get_parser(self, prog_name):
        parser = super(ListLeases, self).get_parser(prog_name)
        parser.add_argument(
            '--sort-by', metavar="<lease_column>",
            help='column name used to sort result',
            default='name'
        )
        parser.add_argument(
            '--project-id', metavar="<project_id>",
            help='ID of the project to filter leases by',
        )
        parser.add_argument(
            '--status', metavar="<status>",
            help='status to filter leases by',
        )
        parser.add_argument(
            '--user', metavar="<user_id>",
            help='User ID to filter leases by',
        )
        return parser

    def get_data(self, parsed_args):
        if parsed_args.project_id:
            self._filters.append(
                lambda x: x['project_id'] == parsed_args.project_id
            )
        if parsed_args.status:
            self._filters.append(
                lambda x: x['status'].lower() == parsed_args.status.lower()
            )
        if parsed_args.user:
            self._filters.append(
                lambda x: x['user_id'] == parsed_args.user
            )
        return super(ListLeases, self).get_data(parsed_args)


class ShowLease(command.ShowCommand):
    """Show details about the given lease."""
    resource = 'lease'
    json_indent = 4
    name_key = 'name'
    log = logging.getLogger(__name__ + '.ShowLease')

    def get_parser(self, prog_name):
        parser = super(ShowLease, self).get_parser(prog_name)
        parser.add_argument(
            '--detail',
            action='store_true',
            help='Return all resources reserved in lease.',
            default=False
        )
        if self.allow_names:
            help_str = 'ID or name of %s to look up'
        else:
            help_str = 'ID of %s to look up'
        parser.add_argument('id', metavar=self.resource.upper(),
                            help=help_str % self.resource)
        return parser

    def args2body(self, parsed_args):
        params = {}
        params['detail'] = parsed_args.detail
        return params


class CreateLeaseBase(command.CreateCommand):
    """Create a lease."""
    resource = 'lease'
    json_indent = 4
    log = logging.getLogger(__name__ + '.CreateLease')
    default_start = 'now'
    default_end = _utc_now() + datetime.timedelta(days=1)

    def get_parser(self, prog_name):
        parser = super(CreateLeaseBase, self).get_parser(prog_name)
        parser.add_argument(
            'name', metavar=self.resource.upper(),
            help='Name for the %s' % self.resource
        )
        parser.add_argument(
            '--start-date',
            dest='start',
            help='Time (YYYY-MM-DD HH:MM) UTC TZ for starting the lease '
                 '(default: current time on the server)',
            default=self.default_start
        )
        parser.add_argument(
            '--end-date',
            dest='end',
            help='Time (YYYY-MM-DD HH:MM) UTC TZ for ending the lease '
                 '(default: 24h from now)',
            default=self.default_end
        )
        parser.add_argument(
            '--before-end-date',
            dest='before_end',
            help='Time (YYYY-MM-DD HH:MM) UTC TZ for taking an action before '
                 'the end of the lease (default: depends on system default)',
            default=None
        )
        parser.add_argument(
            '--reservation',
            metavar="<key=value>",
            action='append',
            dest='reservations',
            help='key/value pairs for creating a generic reservation. '
                 'Specify option multiple times to create multiple '
                 'reservations. ',
            default=[]
        )
        parser.add_argument(
            '--event', metavar='<event_type=str,event_date=time>',
            action='append',
            dest='events',
            help='Creates an event with key/value pairs for the lease. '
                 'Specify option multiple times to create multiple events. '
                 'event_type: type of event (e.g. notification). '
                 'event_date: Time for event (YYYY-MM-DD HH:MM) UTC TZ. ',
            default=[]
        )
        return parser

    def args2body(self, parsed_args):
        params = self._generate_params(parsed_args)
        if not params['reservations']:
            raise exception.IncorrectLease
        return params

    def _generate_params(self, parsed_args):
        params = {}
        if parsed_args.name:
            params['name'] = parsed_args.name
        if not isinstance(parsed_args.start, datetime.datetime):
            if parsed_args.start != 'now':
                try:
                    parsed_args.start = datetime.datetime.strptime(
                        parsed_args.start, '%Y-%m-%d %H:%M')
                except ValueError:
                    raise exception.IncorrectLease
        if not isinstance(parsed_args.end, datetime.datetime):
            try:
                parsed_args.end = datetime.datetime.strptime(
                    parsed_args.end, '%Y-%m-%d %H:%M')
            except ValueError:
                raise exception.IncorrectLease

        if parsed_args.start == 'now':
            start = _utc_now()
        else:
            start = parsed_args.start

        if start > parsed_args.end:
            raise exception.IncorrectLease

        if parsed_args.before_end:
            try:
                parsed_args.before_end = datetime.datetime.strptime(
                    parsed_args.before_end, '%Y-%m-%d %H:%M')
            except ValueError:
                raise exception.IncorrectLease
            if (parsed_args.before_end < start or
                    parsed_args.end < parsed_args.before_end):
                raise exception.IncorrectLease
            params['before_end'] = datetime.datetime.strftime(
                parsed_args.before_end, '%Y-%m-%d %H:%M')

        if parsed_args.start == 'now':
            params['start'] = parsed_args.start
        else:
            params['start'] = datetime.datetime.strftime(parsed_args.start,
                                                         '%Y-%m-%d %H:%M')
        params['end'] = datetime.datetime.strftime(parsed_args.end,
                                                   '%Y-%m-%d %H:%M')

        params['reservations'] = []
        params['events'] = []

        reservations = []
        for res_str in parsed_args.reservations:
            err_msg = ("Invalid reservation argument '%s'. "
                       "Reservation arguments must be of the "
                       "form --reservation <key=value>"
                       % res_str)

            if "physical:host" in res_str:
                defaults = CREATE_RESERVATION_KEYS['physical:host']
            elif "virtual:instance" in res_str:
                defaults = CREATE_RESERVATION_KEYS['virtual:instance']
            elif "virtual:floatingip" in res_str:
                defaults = CREATE_RESERVATION_KEYS['virtual:floatingip']
            elif "network" in res_str:
                defaults = CREATE_RESERVATION_KEYS['network']
            elif "device" in res_str:
                defaults = CREATE_RESERVATION_KEYS['device']
            else:
                defaults = CREATE_RESERVATION_KEYS['others']

            res_info = self._parse_params(res_str, defaults, err_msg)
            reservations.append(res_info)

        if reservations:
            params['reservations'] += reservations

        events = []
        for event_str in parsed_args.events:
            err_msg = ("Invalid event argument '%s'. "
                       "Event arguments must be of the "
                       "form --event <event_type=str,event_date=time>"
                       % event_str)
            event_info = {"event_type": "", "event_date": ""}
            for kv_str in event_str.split(","):
                try:
                    k, v = kv_str.split("=", 1)
                except ValueError:
                    raise exception.IncorrectLease(err_msg)
                if k in event_info:
                    event_info[k] = v
                else:
                    raise exception.IncorrectLease(err_msg)
            if not event_info['event_type'] and not event_info['event_date']:
                raise exception.IncorrectLease(err_msg)
            event_date = event_info['event_date']
            try:
                date = datetime.datetime.strptime(event_date, '%Y-%m-%d %H:%M')
                event_date = datetime.datetime.strftime(date, '%Y-%m-%d %H:%M')
                event_info['event_date'] = event_date
            except ValueError:
                raise exception.IncorrectLease
            events.append(event_info)
        if events:
            params['events'] = events

        return params

    def _parse_params(self, str_params, default, err_msg):
        request_params = {}
        prog = re.compile('^(?:(.*),)?(%s)=(.*)$'
                          % "|".join(default.keys()))

        while str_params != "":
            match = prog.search(str_params)

            if match is None:
                raise exception.IncorrectLease(err_msg)

            self.log.info("Matches: %s", match.groups())
            k, v = match.group(2, 3)
            if k in request_params.keys():
                raise exception.DuplicatedLeaseParameters(err_msg)
            else:
                if strutils.is_int_like(v):
                    request_params[k] = int(v)
                elif isinstance(default[k], list):
                    request_params[k] = jsonutils.loads(v)
                else:
                    request_params[k] = v

            str_params = match.group(1) if match.group(1) else ""

        request_params.update({k: v for k, v in default.items()
                               if k not in request_params.keys() and
                               v is not None})
        return request_params


class CreateLease(CreateLeaseBase):

    def get_parser(self, prog_name):
        parser = super(CreateLease, self).get_parser(prog_name)
        parser.add_argument(
            '--physical-reservation',
            metavar="<min=int,max=int,hypervisor_properties=str,"
                    "resource_properties=str,before_end=str>",
            action='append',
            dest='physical_reservations',
            help='Create a reservation for physical compute hosts. '
                 'Specify option multiple times to create multiple '
                 'reservations. '
                 'min: minimum number of hosts to reserve. '
                 'max: maximum number of hosts to reserve. '
                 'hypervisor_properties: JSON string, see doc. '
                 'resource_properties: JSON string, see doc. '
                 'before_end: JSON string, see doc. ',
            default=[]
        )
        return parser

    def args2body(self, parsed_args):
        params = self._generate_params(parsed_args)

        physical_reservations = []
        for phys_res_str in parsed_args.physical_reservations:
            err_msg = ("Invalid physical-reservation argument '%s'. "
                       "Reservation arguments must be of the "
                       "form --physical-reservation <min=int,max=int,"
                       "hypervisor_properties=str,resource_properties=str,"
                       "before_end=str>"
                       % phys_res_str)
            defaults = CREATE_RESERVATION_KEYS["physical:host"]
            phys_res_info = self._parse_params(phys_res_str, defaults, err_msg)

            if not (phys_res_info['min'] and phys_res_info['max']):
                raise exception.IncorrectLease(err_msg)

            if not (strutils.is_int_like(phys_res_info['min']) and
                    strutils.is_int_like(phys_res_info['max'])):
                raise exception.IncorrectLease(err_msg)

            min_host = int(phys_res_info['min'])
            max_host = int(phys_res_info['max'])

            if min_host > max_host:
                err_msg = ("Invalid physical-reservation argument '%s'. "
                           "Reservation argument min value must be "
                           "less than max value"
                           % phys_res_str)
                raise exception.IncorrectLease(err_msg)

            if min_host == 0 or max_host == 0:
                err_msg = ("Invalid physical-reservation argument '%s'. "
                           "Reservation arguments min and max values "
                           "must be greater than or equal to 1"
                           % phys_res_str)
                raise exception.IncorrectLease(err_msg)

            # NOTE(sbauza): The resource type should be conf-driven mapped with
            #               blazar.conf file but that's potentially on another
            #               host
            phys_res_info['resource_type'] = 'physical:host'
            physical_reservations.append(phys_res_info)
        if physical_reservations:
            # We prepend the physical_reservations to preserve legacy order
            # of reservations
            params['reservations'] = physical_reservations \
                + params['reservations']

        return params


class UpdateLease(command.UpdateCommand):
    """Update a lease."""
    resource = 'lease'
    json_indent = 4
    name_key = 'name'
    log = logging.getLogger(__name__ + '.UpdateLease')

    def get_parser(self, prog_name):
        parser = super(UpdateLease, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            help='New name for the lease',
            default=None
        )
        parser.add_argument(
            '--reservation',
            metavar="<key=value>",
            action='append',
            help='Reservation values to update. The reservation must be '
                 'selected with the id=<reservation-id> key-value pair.',
            default=None)

        # prolong-for and reduce_by are mutually exclusive
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--prolong-for',
            help='Time to prolong lease for',
            default=None
        )
        group.add_argument(
            '--prolong_for',
            help=argparse.SUPPRESS,
            default=None
        )
        group.add_argument(
            '--reduce-by',
            help='Time to reduce lease by',
            default=None
        )
        group.add_argument(
            '--end-date',
            help='end date of the lease',
            default=None)

        # defer-by and a 'future' advance-by are mutually exclusive
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--defer-by',
            help='Time to defer the lease start',
            default=None
        )
        group.add_argument(
            '--advance-by',
            help='Time to advance the lease start',
            default=None
        )
        group.add_argument(
            '--start-date',
            help='start date of the lease',
            default=None)

        return parser

    def args2body(self, parsed_args):
        params = {}
        if parsed_args.name:
            params['name'] = parsed_args.name
        if parsed_args.prolong_for:
            params['prolong_for'] = parsed_args.prolong_for
        if parsed_args.reduce_by:
            params['reduce_by'] = parsed_args.reduce_by
        if parsed_args.end_date:
            params['end_date'] = parsed_args.end_date
        if parsed_args.defer_by:
            params['defer_by'] = parsed_args.defer_by
        if parsed_args.advance_by:
            params['advance_by'] = parsed_args.advance_by
        if parsed_args.start_date:
            params['start_date'] = parsed_args.start_date
        if parsed_args.reservation:
            keys = set([
                # General keys
                'id',
                # Keys for host reservation
                'min', 'max', 'hypervisor_properties', 'resource_properties',
                # Keys for instance reservation
                'vcpus', 'memory_mb', 'disk_gb', 'amount', 'affinity',
                # Keys for floating IP reservation
                'amount', 'network_id', 'required_floatingips',
            ])
            list_keys = ['required_floatingips']
            params['reservations'] = []
            reservations = []
            for res_str in parsed_args.reservation:
                err_msg = ("Invalid reservation argument '%s'. "
                           "Reservation arguments must be of the form "
                           "--reservation <key=value>" % res_str)
                res_info = {}
                prog = re.compile('^(?:(.*),)?(%s)=(.*)$' % '|'.join(keys))

                def parse_params(params):
                    match = prog.search(params)
                    if match:
                        k, v = match.group(2, 3)
                        if k in list_keys:
                            v = jsonutils.loads(v)
                        elif strutils.is_int_like(v):
                            v = int(v)
                        res_info[k] = v
                        if match.group(1) is not None:
                            parse_params(match.group(1))

                parse_params(res_str)
                if res_info:
                    if 'id' not in res_info:
                        raise exception.IncorrectLease(
                            'The key-value pair id=<reservation_id> is '
                            'required for the --reservation argument')
                    reservations.append(res_info)
            if not reservations:
                raise exception.IncorrectLease(err_msg)
            params['reservations'] = reservations
        return params


class DeleteLease(command.DeleteCommand):
    """Delete a lease."""
    resource = 'lease'
    name_key = 'name'
    log = logging.getLogger(__name__ + '.DeleteLease')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from osc_lib import utils


LOG = logging.getLogger(__name__)

DEFAULT_API_VERSION = '1'

# Required by the OSC plugin interface
API_NAME = 'reservation'
API_VERSION_OPTION = 'os_reservations_api_version'
API_VERSIONS = {
    '1': 'blazarclient.v1.client.Client',
}


# Required by the OSC plugin interface
def make_client(instance):
    reservation_client = utils.get_client_class(
        API_NAME,
        instance._api_version[API_NAME],
        API_VERSIONS)

    LOG.debug("Instantiating reservation client: %s", reservation_client)

    client = reservation_client(
        instance._api_version[API_NAME],
        session=instance.session,
        endpoint_override=instance.get_endpoint_for_service_type(
            API_NAME,
            interface=instance.interface,
            region_name=instance._region_name)
    )
    return client


# Required by the OSC plugin interface
def build_option_parser(parser):
    """Hook to add global options.
    Called from openstackclient.shell.OpenStackShell.__init__()
    after the builtin parser has been initialized.  This is
    where a plugin can add global options such as an API version setting.
    :param argparse.ArgumentParser parser: The parser object that has been
        initialized by OpenStackShell.
    """
    parser.add_argument(
        "--os-reservation-api-version",
        metavar="<reservation-api-version>",
        help="Reservation API version, default="
             "{} (Env: OS_RESERVATION_API_VERSION)".format(
                 DEFAULT_API_VERSION)
    )
    return parser



# Copyright (c) 2018 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from unittest import mock

from blazarclient import shell
from blazarclient import tests
from blazarclient.v1.shell_commands import hosts


class CreateHostTest(tests.TestCase):

    def setUp(self):
        super(CreateHostTest, self).setUp()
        self.create_host = hosts.CreateHost(shell.BlazarShell(), mock.Mock())

    def test_args2body(self):
        args = argparse.Namespace(
            name='test-host',
            extra_capabilities=[
                'extra_key1=extra_value1',
                'extra_key2=extra_value2',
            ]
        )

        expected = {
            'name': 'test-host',
            'extra_key1': 'extra_value1',
            'extra_key2': 'extra_value2',
        }

        ret = self.create_host.args2body(args)
        self.assertDictEqual(ret, expected)


class UpdateHostTest(tests.TestCase):

    def create_update_command(self, list_value):
        mock_host_manager = mock.Mock()
        mock_host_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.host = mock_host_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return hosts.UpdateHost(blazar_shell, mock.Mock()), mock_host_manager

    def test_update_host(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        update_host, host_manager = self.create_update_command(list_value)
        args = argparse.Namespace(
            id='101',
            extra_capabilities=[
                'key1=value1',
                'key2=value2'
            ])
        expected = {
            'values': {
                'key1': 'value1',
                'key2': 'value2'
            }
        }
        update_host.run(args)
        host_manager.update.assert_called_once_with('101', **expected)

    def test_update_host_with_name(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        update_host, host_manager = self.create_update_command(list_value)
        args = argparse.Namespace(
            id='host-1',
            extra_capabilities=[
                'key1=value1',
                'key2=value2'
            ])
        expected = {
            'values': {
                'key1': 'value1',
                'key2': 'value2'
            }
        }
        update_host.run(args)
        host_manager.update.assert_called_once_with('101', **expected)

    def test_update_host_with_name_startwith_number(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': '1-host'},
            {'id': '201', 'hypervisor_hostname': '2-host'},
        ]
        update_host, host_manager = self.create_update_command(list_value)
        args = argparse.Namespace(
            id='1-host',
            extra_capabilities=[
                'key1=value1',
                'key2=value2'
            ])
        expected = {
            'values': {
                'key1': 'value1',
                'key2': 'value2'
            }
        }
        update_host.run(args)
        host_manager.update.assert_called_once_with('101', **expected)


class UnsetAttributesHostTest(tests.TestCase):

    def create_unset_command(self, list_value):
        mock_host_manager = mock.Mock()
        mock_host_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.host = mock_host_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return hosts.UnsetAttributesHost(blazar_shell, mock.Mock()), mock_host_manager

    def test_unset_host(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        unset_host, host_manager = self.create_unset_command(list_value)
        extra_caps = ['key1', 'key2']
        args = argparse.Namespace(
            id='101',
            extra_capabilities=extra_caps
        )
        expected = {
            'values': {key: None for key in extra_caps}
        }
        unset_host.run(args)
        host_manager.update.assert_called_once_with('101', **expected)

class ShowHostTest(tests.TestCase):

    def create_show_command(self, list_value, get_value):
        mock_host_manager = mock.Mock()
        mock_host_manager.list.return_value = list_value
        mock_host_manager.get.return_value = get_value

        mock_client = mock.Mock()
        mock_client.host = mock_host_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return hosts.ShowHost(blazar_shell, mock.Mock()), mock_host_manager

    def test_show_host(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        get_value = {
            'id': '101', 'hypervisor_hostname': 'host-1'}

        show_host, host_manager = self.create_show_command(list_value,
                                                           get_value)

        args = argparse.Namespace(id='101', formatter="table")
        expected = [('hypervisor_hostname', 'id'), ('host-1', '101')]

        ret = show_host.get_data(args)
        self.assertEqual(ret, expected)

        host_manager.get.assert_called_once_with('101')

    def test_show_host_with_name(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        get_value = {
            'id': '101', 'hypervisor_hostname': 'host-1'}

        show_host, host_manager = self.create_show_command(list_value,
                                                           get_value)

        args = argparse.Namespace(id='host-1', formatter="table")
        expected = [('hypervisor_hostname', 'id'), ('host-1', '101')]

        ret = show_host.get_data(args)
        self.assertEqual(ret, expected)

        host_manager.get.assert_called_once_with('101')

    def test_show_host_with_name_startwith_number(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': '1-host'},
            {'id': '201', 'hypervisor_hostname': '2-host'},
        ]
        get_value = {
            'id': '101', 'hypervisor_hostname': '1-host'}

        show_host, host_manager = self.create_show_command(list_value,
                                                           get_value)
        args = argparse.Namespace(id='1-host', formatter="table")
        expected = [('hypervisor_hostname', 'id'), ('1-host', '101')]

        ret = show_host.get_data(args)
        self.assertEqual(ret, expected)

        host_manager.get.assert_called_once_with('101')


class DeleteHostTest(tests.TestCase):

    def create_delete_command(self, list_value):
        mock_host_manager = mock.Mock()
        mock_host_manager.list.return_value = list_value

        mock_client = mock.Mock()
        mock_client.host = mock_host_manager

        blazar_shell = shell.BlazarShell()
        blazar_shell.client = mock_client
        return hosts.DeleteHost(blazar_shell, mock.Mock()), mock_host_manager

    def test_delete_host(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        delete_host, host_manager = self.create_delete_command(list_value)

        args = argparse.Namespace(id='101')
        delete_host.run(args)

        host_manager.delete.assert_called_once_with('101')

    def test_delete_host_with_name(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': 'host-1'},
            {'id': '201', 'hypervisor_hostname': 'host-2'},
        ]
        delete_host, host_manager = self.create_delete_command(list_value)

        args = argparse.Namespace(id='host-1')
        delete_host.run(args)

        host_manager.delete.assert_called_once_with('101')

    def test_delete_host_with_name_startwith_number(self):
        list_value = [
            {'id': '101', 'hypervisor_hostname': '1-host'},
            {'id': '201', 'hypervisor_hostname': '2-host'},
        ]
        delete_host, host_manager = self.create_delete_command(list_value)

        args = argparse.Namespace(id='1-host')
        delete_host.run(args)

        host_manager.delete.assert_called_once_with('101')



# Copyright (c) 2019 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from blazarclient import command
from blazarclient import utils

RESOURCE_ID_PATTERN = '^[0-9]+$'


class ShowAllocations(command.ShowCommand):
    """Show allocations for resource identified by type and ID."""
    resource = 'allocation'
    json_indent = 4
    id_pattern = RESOURCE_ID_PATTERN
    name_key = 'hypervisor_hostname'
    log = logging.getLogger(__name__ + '.ShowHostAllocation')

    def get_parser(self, prog_name):
        parser = super(ShowAllocations, self).get_parser(prog_name)
        parser.add_argument(
            'resource_type',
            choices=['host'],
            help='Show allocations for a resource type'
        )
        if self.allow_names:
            help_str = 'ID or name of %s to look up'
        else:
            help_str = 'ID of %s to look up'
        parser.add_argument('id', metavar="RESOURCE",
                            help=help_str % "resource")
        parser.add_argument(
            '--reservation-id',
            dest='reservation_id',
            default=None,
            help='Show only allocations with specific reservation_id'
        )
        parser.add_argument(
            '--lease-id',
            dest='lease_id',
            default=None,
            help='Show only allocations with specific lease_id'
        )
        return parser

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        blazar_client = self.get_client()
        resource_manager = getattr(blazar_client, self.resource)

        if self.allow_names:
            res_id = utils.find_resource_id_by_name_or_id(
                blazar_client,
                parsed_args.resource_type,
                parsed_args.id,
                self.name_key,
                self.id_pattern)
        else:
            res_id = parsed_args.id

        data = resource_manager.get(
            self.args2body(parsed_args)['resource'], res_id)

        if parsed_args.lease_id is not None:
            data['reservations'] = list(
                filter(lambda d: d['lease_id'] == parsed_args.lease_id,
                       data['reservations']))
        if parsed_args.reservation_id is not None:
            data['reservations'] = list(
                filter(lambda d: d['id'] == parsed_args.reservation_id,
                       data['reservations']))

        return list(zip(*sorted(data.items())))

    def args2body(self, parsed_args):
        params = {}
        if parsed_args.resource_type == 'host':
            params.update(dict(resource='os-hosts'))
        return params


class ListAllocations(command.ListCommand):
    """List allocations for all resources of a type."""
    resource = 'allocation'
    log = logging.getLogger(__name__ + '.ListHostAllocations')
    list_columns = ['resource_id', 'reservations']

    def get_parser(self, prog_name):
        parser = super(ListAllocations, self).get_parser(prog_name)

        parser.add_argument(
            'resource_type',
            choices=['host'],
            help='Show allocations for a resource type'
        )
        parser.add_argument(
            '--reservation-id',
            dest='reservation_id',
            default=None,
            help='Show only allocations with specific reservation_id'
        )
        parser.add_argument(
            '--lease-id',
            dest='lease_id',
            default=None,
            help='Show only allocations with specific lease_id'
        )
        parser.add_argument(
            '--sort-by', metavar="<allocation_column>",
            help='column name used to sort result',
            default='resource_id'
        )
        return parser

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        data = self.retrieve_list(parsed_args)

        for resource in data:
            if parsed_args.lease_id is not None:
                resource['reservations'] = list(
                    filter(lambda d: d['lease_id'] == parsed_args.lease_id,
                           resource['reservations']))
            if parsed_args.reservation_id is not None:
                resource['reservations'] = list(
                    filter(lambda d: d['id'] == parsed_args.reservation_id,
                           resource['reservations']))

        return self.setup_columns(data, parsed_args)

    def args2body(self, parsed_args):
        params = super(ListAllocations, self).args2body(parsed_args)
        if parsed_args.resource_type == 'host':
            params.update(dict(resource='os-hosts'))
        return params



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os
import re

from oslo_serialization import jsonutils as json

from blazarclient import exception
from blazarclient.i18n import _

ELAPSED_TIME_REGEX = '^(\d+)([s|m|h|d])$' # noqa W605

LEASE_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'
API_DATE_FORMAT = '%Y-%m-%d %H:%M'


def env(*args, **kwargs):
    """Returns the first environment variable set.

    if none are non-empty, defaults to '' or keyword arg default.
    """
    for v in args:
        value = os.environ.get(v)
        if value:
            return value
    return kwargs.get('default', '')


def to_primitive(value):
    if isinstance(value, list) or isinstance(value, tuple):
        o = []
        for v in value:
            o.append(to_primitive(v))
        return o
    elif isinstance(value, dict):
        o = {}
        for k, v in value.items():
            o[k] = to_primitive(v)
        return o
    elif isinstance(value, datetime.datetime):
        return str(value)
    elif hasattr(value, 'iteritems'):
        return to_primitive(dict(value.items()))
    elif hasattr(value, '__iter__'):
        return to_primitive(list(value))
    else:
        return value


def dumps(value, indent=None):
    try:
        return json.dumps(value, indent=indent)
    except TypeError:
        pass
    return json.dumps(to_primitive(value))


def get_item_properties(item, fields, mixed_case_fields=None, formatters=None):
    """Return a tuple containing the item properties.

    :param item: a single item resource (e.g. Server, Tenant, etc)
    :param fields: tuple of strings with the desired field names
    :param mixed_case_fields: tuple of field names to preserve case
    :param formatters: dictionary mapping field names to callables
       to format the values
    """
    row = []
    if mixed_case_fields is None:
        mixed_case_fields = []
    if formatters is None:
        formatters = {}

    for field in fields:
        if field in formatters:
            row.append(formatters[field](item))
        else:
            if field in mixed_case_fields:
                field_name = field.replace(' ', '_')
            else:
                field_name = field.lower().replace(' ', '_')
            if not hasattr(item, field_name) and isinstance(item, dict):
                # display empty string if the resource does not have a field
                data = item.get(field_name, '')
            else:
                data = getattr(item, field_name, '')
            if data is None:
                data = ''
            row.append(data)
    return tuple(row)


def find_resource_id_by_name_or_id(client, resource_type, name_or_id,
                                   name_key, id_pattern):
    try:
        # Since IDs and Hypervisor Hostnames can both be UUIDs, we'll first
        # try to look up by Hypervisor Hostname
        return _find_resource_id_by_name(client, resource_type,
                                         name_or_id, name_key)
    except exception.BlazarClientException as e:
        # If we fail to find a host by Hypervisor Hostname, we'll check if the
        # key could be an ID, and then we'll return that.
        if re.match(id_pattern, name_or_id):
            return name_or_id
        else:
            raise e


def _find_resource_id_by_name(client, resource_type, name, name_key):
    resource_manager = getattr(client, resource_type)
    resources = resource_manager.list()

    named_resources = []
    key = name_key if name_key else 'name'
    for resource in resources:
        if str(resource.get(key)) == name:
            named_resources.append(resource['id'])
    if len(named_resources) > 1:
        raise exception.NoUniqueMatch(message="There are more than one "
                                              "appropriate resources for the "
                                              "name '%s' and type '%s'" %
                                              (name, resource_type))
    elif named_resources:
        return named_resources[0]
    else:
        message = "Unable to find resource with name '%s'" % name
        raise exception.BlazarClientException(message=message,
                                              status_code=404)


def from_elapsed_time_to_seconds(elapsed_time, pos_sign=True):
    """Return the positive or negative amount of seconds based on the
    elapsed_time parameter with a sign depending on the sign parameter.
    :param: elapsed_time: a string that matches ELAPSED_TIME_REGEX
    :param: sign: if pos_sign is True, the returned value will be positive.
    Otherwise it will be positive.
    """
    is_elapsed_time = re.match(ELAPSED_TIME_REGEX, elapsed_time)
    if is_elapsed_time is None:
        raise exception.BlazarClientException(_("Invalid time "
                                                "format for option."))
    elapsed_time_value = int(is_elapsed_time.group(1))
    elapsed_time_option = is_elapsed_time.group(2)
    seconds = {
        's': lambda x:
        datetime.timedelta(seconds=x).total_seconds(),
        'm': lambda x:
        datetime.timedelta(minutes=x).total_seconds(),
        'h': lambda x:
        datetime.timedelta(hours=x).total_seconds(),
        'd': lambda x:
        datetime.timedelta(days=x).total_seconds(),
    }[elapsed_time_option](elapsed_time_value)

    # the above code returns a "float"
    if pos_sign:
        return int(seconds)
    return int(seconds) * -1


def from_elapsed_time_to_delta(elapsed_time, pos_sign=True):
    """Return the positive or negative delta time based on the
    elapsed_time parameter.
    :param: elapsed_time: a string that matches ELAPSED_TIME_REGEX
    :param: sign: if sign is True, the returned value will be negative.
    Otherwise it will be positive.
    """
    seconds = from_elapsed_time_to_seconds(elapsed_time, pos_sign=pos_sign)
    return datetime.timedelta(seconds=seconds)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ast
import logging

from cliff import command
from cliff.formatters import table
from cliff import lister
from cliff import show

from blazarclient import exception
from blazarclient import utils

HEX_ELEM = '[0-9A-Fa-f]'
UUID_PATTERN = '-'.join([HEX_ELEM + '{8}', HEX_ELEM + '{4}',
                         HEX_ELEM + '{4}', HEX_ELEM + '{4}',
                         HEX_ELEM + '{12}'])


class OpenStackCommand(command.Command):
    """Base class for OpenStack commands."""

    api = None

    def run(self, parsed_args):
        if not self.api:
            return
        else:
            return super(OpenStackCommand, self).run(parsed_args)

    def get_data(self, parsed_args):
        pass

    def take_action(self, parsed_args):
        return self.get_data(parsed_args)


class TableFormatter(table.TableFormatter):
    """This class is used to keep consistency with prettytable 0.6."""

    def emit_list(self, column_names, data, stdout, parsed_args):
        if column_names:
            super(TableFormatter, self).emit_list(column_names, data, stdout,
                                                  parsed_args)
        else:
            stdout.write('\n')


class BlazarCommand(OpenStackCommand):

    """Base Blazar CLI command."""
    api = 'reservation'
    log = logging.getLogger(__name__ + '.BlazarCommand')
    values_specs = []
    json_indent = None
    resource = None
    allow_names = True
    name_key = None
    id_pattern = UUID_PATTERN

    def __init__(self, app, app_args):
        super(BlazarCommand, self).__init__(app, app_args)

        # NOTE(dbelova): This is no longer supported in cliff version 1.5.2
        # the same moment occurred in Neutron:
        # see https://bugs.launchpad.net/python-neutronclient/+bug/1265926

        # if hasattr(self, 'formatters'):
        #     self.formatters['table'] = TableFormatter()

    def get_client(self):
        # client_manager.reservation is used for osc_lib, and should be used
        # if it exists
        if hasattr(self.app, 'client_manager'):
            return self.app.client_manager.reservation
        else:
            return self.app.client

    def get_parser(self, prog_name):
        parser = super(BlazarCommand, self).get_parser(prog_name)
        return parser

    def format_output_data(self, data, parsed_args):
        # Do not format output data if the formatter is not table
        if parsed_args.formatter != 'table':
            return
        for k, v in data.items():
            if isinstance(v, str):
                try:
                    # Deserialize if possible into dict, lists, tuples...
                    v = ast.literal_eval(v)
                except SyntaxError:
                    # NOTE(sbauza): This is probably a datetime string, we need
                    #               to keep it unchanged.
                    pass
                except ValueError:
                    # NOTE(sbauza): This is not something AST can evaluate,
                    #               probably a string.
                    pass
            elif isinstance(v, list) or isinstance(v, dict):
                value = utils.dumps(v, indent=self.json_indent)
                data[k] = value
            elif v is None:
                data[k] = ''

    def add_known_arguments(self, parser):
        pass

    def args2body(self, parsed_args):
        return {}


class CreateCommand(BlazarCommand, show.ShowOne):
    """Create resource with passed args."""

    api = 'reservation'
    resource = None
    log = None

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.create(**body)
        self.format_output_data(data, parsed_args)

        if data:
            print('Created a new %s:' % self.resource, file=self.app.stdout)
        else:
            data = {'': ''}
        return list(zip(*sorted(data.items())))


class UpdateCommand(BlazarCommand):
    """Update resource's information."""

    api = 'reservation'
    resource = None
    log = None

    def get_parser(self, prog_name):
        parser = super(UpdateCommand, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'ID or name of %s to update'
        else:
            help_str = 'ID of %s to update'
        parser.add_argument(
            'id', metavar=self.resource.upper(),
            help=help_str % self.resource
        )
        self.add_known_arguments(parser)
        return parser

    def run(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        if self.allow_names:
            res_id = utils.find_resource_id_by_name_or_id(blazar_client,
                                                          self.resource,
                                                          parsed_args.id,
                                                          self.name_key,
                                                          self.id_pattern)
        else:
            res_id = parsed_args.id
        resource_manager = getattr(blazar_client, self.resource)
        resource_manager.update(res_id, **body)
        print('Updated %s: %s' % (self.resource, parsed_args.id),
              file=self.app.stdout)
        return


class DeleteCommand(BlazarCommand):
    """Delete a given resource."""

    api = 'reservation'
    resource = None
    log = None

    def get_parser(self, prog_name):
        parser = super(DeleteCommand, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'ID or name of %s to delete'
        else:
            help_str = 'ID of %s to delete'
        parser.add_argument(
            'id', metavar=self.resource.upper(),
            help=help_str % self.resource)
        return parser

    def run(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        blazar_client = self.get_client()
        resource_manager = getattr(blazar_client, self.resource)
        if self.allow_names:
            res_id = utils.find_resource_id_by_name_or_id(blazar_client,
                                                          self.resource,
                                                          parsed_args.id,
                                                          self.name_key,
                                                          self.id_pattern)
        else:
            res_id = parsed_args.id
        resource_manager.delete(res_id)
        print('Deleted %s: %s' % (self.resource, parsed_args.id),
              file=self.app.stdout)
        return


class ListCommand(BlazarCommand, lister.Lister):
    """List resources that belong to a given tenant."""

    api = 'reservation'
    resource = None
    log = None
    _formatters = {}
    list_columns = []
    long_columns = []
    unknown_parts_flag = True
    _filters = []

    def args2body(self, parsed_args):
        params = {}
        if parsed_args.sort_by:
            if parsed_args.sort_by in self.list_columns:
                params['sort_by'] = parsed_args.sort_by
            elif self.long_columns and parsed_args.sort_by in self.long_columns:
                params['sort_by'] = parsed_args.sort_by
            else:
                msg = 'Invalid sort option %s' % parsed_args.sort_by
                raise exception.BlazarClientException(msg)
        return params

    def get_parser(self, prog_name):
        parser = super(ListCommand, self).get_parser(prog_name)
        if self.long_columns:
            parser.add_argument(
                '--long',
                action='store_true',
                help='Display detailed information for each item'
            )
        return parser

    def retrieve_list(self, parsed_args):
        """Retrieve a list of resources from Blazar server."""
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.list(**body)
        for f in self._filters:
            data = list(filter(f, data))
        return data

    def setup_columns(self, info, parsed_args):
        """
        Determines the list of columns that may be visible to the client. This may not
        be the columns that are actually visible to the client on the output of their
        command. The output columns are determined by a process in cliff.display which
        compares the parsed_args to the list output by this function.
        """
        # return empty list when info is empty
        # or sort all keys that are in all the networks
        columns = (
            len(info) > 0 and sorted({k for d in info for k in d.keys()}) or []
        )
        if parsed_args.columns:
            valid_parsed_columns = {col for col in parsed_args.columns if col in columns}
        else:
            valid_parsed_columns = set()

        default_columns = self.list_columns
        if self.long_columns and parsed_args.long:
            default_columns += self.long_columns
        if default_columns:
            columns = {
                col for col in default_columns if col in columns
            } | valid_parsed_columns
            # sort the columns based on list_columns
            sorting_map = {item: i for i, item in enumerate(default_columns)}
            # sort key is either index of item in list_columns, or placed at end of list
            columns = sorted(columns, key=lambda x: sorting_map.get(x, len(default_columns)))

        return (
            columns,
            (utils.get_item_properties(s, columns, formatters=self._formatters)
             for s in info)
        )

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        data = self.retrieve_list(parsed_args)
        return self.setup_columns(data, parsed_args)


class ListAllocationCommand(ListCommand, lister.Lister):
    """List allocations that belong to a given tenant."""

    def get_parser(self, prog_name):
        parser = super(ListAllocationCommand, self).get_parser(prog_name)
        parser.add_argument(
            '--lease-id', metavar="<lease_id>",
            help='Lease to filter allocations by',
        )
        parser.add_argument(
            '--username', metavar="<username>",
            help='Username to filter allocations by',
        )
        return parser

    def retrieve_list(self, parsed_args):
        """Retrieve a list of resources from Blazar server."""
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.list_allocations(**body)
        filters = []
        if parsed_args.lease_id:
            filters.append(lambda x: x['lease_id'] == parsed_args.lease_id)
        if parsed_args.username:
            filters.append(lambda x: x.get("extras", {}).get('user_name') == parsed_args.username)

        if filters:
            new_data = []
            for allocation in data:
                new_reservations = []
                for reservation in allocation["reservations"]:
                    if all(f(reservation) for f in filters):
                        new_reservations.append(reservation)
                if new_reservations:
                    new_data.append({
                        "resource_id": allocation["resource_id"],
                        "reservations": new_reservations,
                    })
            data = new_data
        return data


class ShowCommand(BlazarCommand, show.ShowOne):
    """Show information of a given resource."""

    api = 'reservation'
    resource = None
    log = None

    def get_parser(self, prog_name):
        parser = super(ShowCommand, self).get_parser(prog_name)
        return parser

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        body = self.args2body(parsed_args)
        blazar_client = self.get_client()

        if self.allow_names:
            res_id = utils.find_resource_id_by_name_or_id(blazar_client,
                                                          self.resource,
                                                          parsed_args.id,
                                                          self.name_key,
                                                          self.id_pattern)
        else:
            res_id = parsed_args.id

        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.get(res_id)
        if body.get('detail'):
            data.update(resource_manager.additional_details(res_id))
        self.format_output_data(data, parsed_args)
        return list(zip(*sorted(data.items())))


class ShowAllocationCommand(ShowCommand, show.ShowOne):
    """Show allocations for a given resource."""

    def get_parser(self, prog_name):
        parser = super(ShowAllocationCommand, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'ID or name of %s to show allocations'
        else:
            help_str = 'ID of %s to show allocations'
        parser.add_argument(
            'id', metavar=self.resource.upper(),
            help=help_str % self.resource)
        return parser

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        blazar_client = self.get_client()
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.get_allocation(parsed_args.id)
        self.format_output_data(data, parsed_args)
        return list(zip(*sorted(data.items())))


class ReallocateCommand(BlazarCommand):
    """Reallocate host from current leases."""

    api = 'reservation'
    resource = None
    log = None

    def get_parser(self, prog_name):
        parser = super(ReallocateCommand, self).get_parser(prog_name)
        if self.allow_names:
            help_str = 'ID or name of %s to update'
        else:
            help_str = 'ID of %s to update'
        parser.add_argument(
            'id', metavar=self.resource.upper(),
            help=help_str % self.resource
        )
        self.add_known_arguments(parser)
        return parser

    def run(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        if self.allow_names:
            res_id = utils.find_resource_id_by_name_or_id(blazar_client,
                                                          self.resource,
                                                          parsed_args.id,
                                                          self.name_key,
                                                          self.id_pattern)
        else:
            res_id = parsed_args.id
        resource_manager = getattr(blazar_client, self.resource)
        resource_manager.reallocate(res_id, body)
        print('Reallocated %s: %s' % (self.resource, parsed_args.id),
              file=self.app.stdout)
        return


class ShowPropertyCommand(BlazarCommand, show.ShowOne):
    """Show information of a given resource."""

    api = 'reservation'
    resource = None
    log = None

    def get_parser(self, prog_name):
        parser = super(ShowPropertyCommand, self).get_parser(prog_name)
        parser.add_argument('property_name', metavar='PROPERTY_NAME',
                            help='Name of property.')
        return parser

    def get_data(self, parsed_args):
        self.log.debug('get_data(%s)' % parsed_args)
        blazar_client = self.get_client()
        resource_manager = getattr(blazar_client, self.resource)
        data = resource_manager.get_property(parsed_args.property_name)
        return list(zip(*sorted(data.items())))


class UpdatePropertyCommand(BlazarCommand):
    api = 'reservation'
    resource = None
    log = None

    def run(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        blazar_client = self.get_client()
        body = self.args2body(parsed_args)
        resource_manager = getattr(blazar_client, self.resource)
        resource_manager.set_property(**body)
        print(
            'Updated %s property: %s' % (
                self.resource, parsed_args.property_name),
            file=self.app.stdout)
        return

    def get_parser(self, prog_name):
        parser = super(UpdatePropertyCommand, self).get_parser(prog_name)
        parser.add_argument(
            'property_name', metavar='PROPERTY_NAME',
            help='Name of property to patch.'
        )
        parser.add_argument(
            '--private',
            action='store_true',
            default=False,
            help='Set property to private.'
        )
        parser.add_argument(
            '--public',
            action='store_true',
            default=False,
            help='Set property to public.'
        )
        parser.add_argument(
            '--unique',
            action='store_true',
            default=False,
            help='Set capability as unique.'
        )
        return parser

    def args2body(self, parsed_args):
        return dict(
            property_name=parsed_args.property_name,
            is_unique=(parsed_args.unique is True),
            private=(parsed_args.private is True))



# Copyright (c) 2019 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazarclient import base


class FloatingIPClientManager(base.BaseClientManager):
    """Manager for floating IP requests."""

    def create(self, network_id, floating_ip_address, **kwargs):
        """Creates a floating IP from values passed."""
        values = {'floating_network_id': network_id,
                  'floating_ip_address': floating_ip_address}
        values.update(**kwargs)
        resp, body = self.request_manager.post('/floatingips', body=values)
        return body['floatingip']

    def get(self, floatingip_id):
        """Show floating IP details."""
        resp, body = self.request_manager.get(
            '/floatingips/%s' % floatingip_id)
        return body['floatingip']

    def delete(self, floatingip_id):
        """Deletes floating IP with specified ID."""
        resp, body = self.request_manager.delete(
            '/floatingips/%s' % floatingip_id)

    def list(self, sort_by=None):
        """List all floating IPs."""
        resp, body = self.request_manager.get('/floatingips')
        floatingips = body['floatingips']
        if sort_by:
            floatingips = sorted(floatingips, key=lambda ip: ip[sort_by])
        return floatingips



# Copyright (c) 2019 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazarclient import base


class AllocationClientManager(base.BaseClientManager):
    """Manager for the ComputeHost connected requests."""

    def get(self, resource, resource_id):
        """Get allocation for resource identified by type and ID."""
        resp, body = self.request_manager.get(
            '/%s/%s/allocation' % (resource, resource_id))
        return body['allocation']

    def list(self, resource, sort_by=None):
        """List allocations for all resources of a type."""
        resp, body = self.request_manager.get('/%s/allocations' % resource)
        allocations = body['allocations']
        if sort_by:
            allocations = sorted(allocations, key=lambda alloc: alloc[sort_by])
        return allocations



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import importutils

from blazarclient import exception
from blazarclient.i18n import _


def Client(version=1, service_type='reservation', *args, **kwargs):
    version_map = {
        '1': 'blazarclient.v1.client.Client',
        '1a0': 'blazarclient.v1.client.Client',
    }
    try:
        client_path = version_map[str(version)]
    except (KeyError, ValueError):
        msg = _("Invalid client version '%(version)s'. "
                "Must be one of: %(available_version)s") % ({
                    'version': version,
                    'available_version': ', '.join(version_map.keys())
                })
        raise exception.UnsupportedVersion(msg)

    return importutils.import_object(client_path,
                                     service_type=service_type,
                                     *args, **kwargs)
